import numpy as np
import pandas as pd
from collections import namedtuple


__VER__ = '0.2.0.0'


###
### PLOT UTILS
###


def plot_autoreg(np_train, 
                 np_test, 
                 np_forecast,
                 title,
                 forecast_title='LumAR',
                 subtitles=None,
                 other_forecasts=[],
                 other_forecasts_titles=[],
                 save_with_log=None,
                 save_and_show=True,
                 log_y=False,
                 hist_vlines=None,
                 figsize=(21,13),
                 hspace=0.15,
                 ma_history=None,
                 ma_target=None
                 ):
  """
  Plot time-series with backlog, forecasted values and reality

  Parameters
  ----------
  np_train : ndarray
    the backlog data (training data).
    
  np_test : ndarray
    the test/reality of the forecast period.
    
  np_forecast : ndarray
    the predictions.
    
  title : str
    name of the whole graph.
    
  subtitles : list[str], optional
    name of each sub-plot. The default is None.
    
  other_forecasts : list of ndarrays, optional
    if we want to compare with another auto-regressors we can put here
    its predictions. The default is [].

  other_forecasts_titles: list, optional
    title for each other forecast.
    The default is [].
    
  save_with_log : Logger, optional
    the Logger object that will save the plot if supplied. The default is None.
    
  log_y: bool, optional,
    log y axis if True, default is False
    
  hist_vlines: list[list], optional
    where to draw vertical lines on x axis (for historical points in autoregs) in each 
    individual subgraph plot. Each point x in each list of lines represent -x point relative
    from the end of history. Default is None

  hspace: float, optional
    the size in fig percent of the space between subplots
    
  figsize: tuple floats, optional
    default width and height (height for each subgraph)

  ma_history: ...
    ...

  ma_target: ...
    ...

  Raises
  ------
  ValueError
    DESCRIPTION.

  Returns
  -------
  None.

  """

  static_list_colors_other_forecasts = [
    'c--', 'm--', 'p--', 'y--', 'k--'
  ]

  import matplotlib.pyplot as plt
  plt.style.use('ggplot')
  
  if type(np_train) != np.ndarray:
    raise ValueError("`np_train` must be ndarray")
  if type(np_test) != np.ndarray:
    raise ValueError("`np_test` must be ndarray")
  if type(np_forecast) != np.ndarray:
    raise ValueError("`np_forecast` must be ndarray")

  np_train = np_train.squeeze()
  np_test = np_test.squeeze()
  np_forecast = np_forecast.squeeze()  
  if len(np_train.shape) == 1:
    np_train = np_train.reshape(1,-1)
    np_test = np_test.reshape(1,-1)
    np_forecast = np_forecast.reshape(1, -1)    

  assert np_test.shape == np_forecast.shape, "test and forecasts must have same shape"
  assert np_train.shape[0] == np_test.shape[0], "train and test have different no of series"
  assert type(other_forecasts) is list

  if len(other_forecasts) > 0:
    for _np_forecast in other_forecasts:
      assert _np_forecast.shape == np_test.shape, "other forecast must match test data"
    
  if subtitles is not None and len(subtitles) != np_train.shape[0]:
    raise ValueError("Must have 1 subtitle per subgraph ({} != {}".format(subtitles, np_train.shape[0]))


  hist_len = np_train.shape[1]
  test_len = np_test.shape[1]  
  full_len = hist_len + test_len
  n_series = np_train.shape[0]

  np_train_show = np_train.copy()
  np_test_show = np_test.copy()
  np_forecast_show = np_forecast.copy()
  other_forecasts_show = other_forecasts

  if len(other_forecasts_titles) == 0:
    other_forecasts_titles = ['OthAR{}'.format(idx_other_f)
                              for idx_other_f in range(len(other_forecasts_show))]
  #endif

  other_forecasts_titles = list(map(lambda x: str(test_len)+'-steps-{}'.format(x), other_forecasts_titles))
  forecast_title = str(test_len)+'-steps-{}'.format(forecast_title)

  assert len(other_forecasts_titles) == len(other_forecasts_show)

  EPS_LOG = 0.5
  
  if log_y:
    np_train_show = np.log(np_train_show + EPS_LOG)
    np_test_show = np.log(np_test_show + EPS_LOG)
    np_forecast_show = np.log(np_forecast_show + EPS_LOG)
    other_forecasts_show = list(map(lambda x: np.log(x + EPS_LOG), other_forecasts_show))
  #endif

  FIG_W, FIG_H = figsize
  
  fig, axes =  plt.subplots(n_series,1, figsize=(FIG_W,FIG_H * n_series))
  if type(axes) not in [list, np.ndarray]:
    axes = [axes]
  for i in range(n_series):
    axes[i].plot(np.arange(hist_len), np_train_show[i], 'k-', label='real past')
    axes[i].plot(
      np.arange(hist_len, full_len), 
      np_test_show[i], 
      'r-', 
      linewidth=1.0,
      label='real future')
    axes[i].plot(
      np.arange(hist_len, full_len), np_forecast_show[i], 'b-',
      linewidth=2.0,
      label=forecast_title
      )
    if log_y:
      axes[i].set_ylabel('Log scale')
    if len(other_forecasts_show) > 0:
      for idx_other_f, np_other_forecast_show in enumerate(other_forecasts_show):
        axes[i].plot(
          np.arange(hist_len, full_len),
          np_other_forecast_show[i],
          '{}'.format(static_list_colors_other_forecasts[idx_other_f]),
          label=other_forecasts_titles[idx_other_f],
          linewidth=3.0,
          )
    _ymin = min(np_train[i].min(), np_forecast_show[i].min())
    _ymax = max(np_train[i].max(), np_forecast_show[i].max())
    axes[i].vlines(hist_len, ymin=_ymin, ymax=_ymax, linestyle='--', colors='g')
    if hist_vlines is not None:
      if len(hist_vlines) > i:
        ser_vlines = hist_vlines[i]
        for pnt in ser_vlines:
          axes[i].vlines(hist_len - pnt, ymin=_ymin, ymax=_ymax, linestyle='--', colors='k')
          axes[i].text(
            hist_len - pnt, np_train[i].max(), 
            str(- pnt), 
            ha="center", va="center",
            rotation=45
            )
          # axes[i].annotate(str(- pnt), (hist_len - pnt, np_train[i].max()))
    # axes[i].annotate(str(hist_len), (hist_len, np_train[i].max()))
    axes[i].text(
      hist_len, np_train[i].max(), str(0), 
      ha="center", va="center",
      rotation=45
      )
    if subtitles is not None:
      _subtitle = subtitles[i]
    else:
      _subtitle = 'Series {}/{}'.format(i+1, n_series)
    if '\n' not in _subtitle:
      _subtitle += '\n Total forecast={:.2f},  reality={:.2f}, max_hist={:.2f}'.format(
        np_forecast[i].sum(),
        np_test[i].sum(),
        np_train[i].max()
        )
    axes[i].set_title(_subtitle)
    axes[i].legend(loc='center left')
  fig.suptitle(title + '\n', fontsize=22)  
  if save_with_log is not None:
    save_with_log.add_copyright_to_plot(plt)
  else:
    plt.annotate(
      "ts v{} local".format(__VER__),
      xy=(1, -0.07), xytext=(0, 2),
      xycoords=('axes fraction'),
      textcoords='offset points',
      size=9, ha='right', va='top',   
      bbox=dict(boxstyle="round,pad=0.3", fc="lightgray", ec="lightgray", lw=1),
      )
  plt.tight_layout()
  offset = max(0,(3-n_series) * 0.06)
  fig.subplots_adjust(top=0.95 - offset, hspace=hspace)
  if save_with_log is not None:
    title_fn = title.replace(' ','_').replace('.','_').replace('%','').replace(':','')
    save_with_log.save_plot(plt, label=title_fn, include_prefix=False)
    if not save_and_show:
      plt.close()
      return
  plt.show()
  return
  

def plot_ts(np_values, ma=[7, 14, 30], trend_window=None):
  print("WARNING: `plot_ts` is obsolete. please use `plot_ts_ma`")
  plot_ts_ma(np_values, ma, trend_window)
  
def plot_ts_ma(np_values, ma=[7, 14, 30], trend_window=None):
  import matplotlib.pyplot as plt
  plt.style.use('ggplot')
  
  styles = ['b-', 'g-','k-']
  assert len(ma) <= 3
  n_obs = np_values.shape[0]
  plt.figure(figsize=(13,8)) # 8x5
  plt.plot(np_values, 'r-', label='signal')
  for i, _ma in enumerate(ma):
    if n_obs < _ma:
      continue    
    mavals = moving_averages(np_values, window=_ma)
    plt.plot(mavals, styles[i], label='ma-'+str(_ma))
  plt.title("Timeseries analysis")
  
  if trend_window is None:
    trend_len = min(np_values.shape[0], max(ma))
  else:
    assert trend_window < n_obs
    trend_len = trend_window
  
  # MUST PUT THE TREND WHERE IT SHOULD BE
  # must start from history not from first item
  np_trend_vals = get_series_trends(
    np_values.reshape(1,-1), 
    last_steps=trend_len)
  np_trend_ticks = np.arange(n_obs - trend_len, n_obs)
  np_trend_vals = np_trend_vals.squeeze()[-trend_len:]
  plt.plot(np_trend_ticks, np_trend_vals, 'k--', label='last_{}_trend'.format(trend_len))  
  plt.legend()
  plt.annotate(
    "local",
    xy=(1, -0.07), xytext=(0, 2),
    xycoords=('axes fraction'),
    textcoords='offset points',
    size=9, ha='right', va='top',   
    bbox=dict(boxstyle="round,pad=0.3", fc="lightgray", ec="lightgray", lw=1),
    )

  plt.show()
  return


###
### ARIMA UTILS
###


def get_max_lag(data):
  n_obs = len(data)
  n_ser = 1
  # max_k = np.floor(10 * (np.log10(n_obs) - np.log10(n_ser))).astype(int)
  max_k = n_obs    
  return max_k
  

def acf(data, k, max_len=None):
  """
  Autocorrelation function

  :param data: time series
  :param k: lag
  :return:
  """
  m = np.mean(data)
  s1 = 0
  if max_len is None:
    max_len = len(data)
  assert k < max_len
  for i in range(k, max_len):
      s1 = s1 + ((data[i] - m) * (data[i - k] - m))

  s2 = 0
  for i in range(0, max_len):
      s2 = s2 + ((data[i] - m) ** 2)

  return float(s1 / s2)

def acf_limit(acfs, series_len, k):
  """  
  Returns confidence interval for lag k based on precalculated ACFs up to k

  Parameters
  ----------
  acfs : TYPE
    list or ndarray with precalculated acfs
  series_len : TYPE
    len of series
  k : TYPE
    lag value

  Returns
  -------
  float - confidence value (to be compared with acf @ k)    

  """
  if k==0:
    return 0
  s = acfs[1]
  for i in range(2, k):
      s = s + (acfs[i] ** 2)
  limit = 1.645 * (np.sqrt((1 + 2 * s) / series_len))  
  return limit


def acf_plot(data, max_k=None, plot=True, seas_as_lag=False, max_len=None):
  return calc_acfs(
    np_data=data,
    max_k=max_k,
    plot=plot,
    seas_as_lag=seas_as_lag,
    max_len=max_len,
    )

def calc_acfs(np_data, max_k=None, plot=False, seas_as_lag=False, max_len=None):
  """
  Plots and returns auto-correlation function and 
  returns values, confidence intervals and seasonality signals

  Parameters
  ----------
  data : ndarray
    1d time-series
  max_k : int, optional
    max lag. The default is None.
  plot : bool, optional
    generate pyplot, default True
  seas_as_lag : bool, optional (False)
    will output `seas` as lag values

  Returns
  -------
    tuple(3): acfs, confs, seas:
      acfs: values of autocorr functions for lag=1...
      confs: conf values for autocorr lag=1...
      seas: bools or valid lag values 


  """
  if plot:
    import matplotlib.pyplot as plt
    plt.style.use('ggplot')
  
  if max_len is not None and max_len < len(np_data):
    data = np_data[-max_len:]
  else:
    data = np_data
    
  n_obs = len(data)

  if max_k is None:
    max_k = get_max_lag(data)
    
  
  acfs = []
  conf = []
  for k in range(max_k):
    acfs.append(acf(data, k, max_len=max_len))
    conf.append(acf_limit(acfs=acfs, series_len=n_obs, k=k))

  np_conf = np.array(conf)    
  np_acfs = np.array(acfs)
  np_season = np.abs(np_acfs) > np_conf
    
  if plot:
    np_k = np.arange(max_k)
    np_bad_k = np_k[~np_season]
    np_bad_acf = np_acfs[~np_season]
    np_good_k = np_k[np_season]
    np_good_acf = np_acfs[np_season]
    plt.figure(figsize=(13,8))
    plt.vlines(range(max_k),[0],acfs)
    plt.plot(np_bad_k, np_bad_acf, "bo")
    plt.plot(np_good_k, np_good_acf, "go")
    plt.fill_between(range(1, max_k), y1=np_conf[1:], y2=-np_conf[1:], alpha=0.25, label='low conf area')  
    _ymin = -1
    _ymax = 1
    plt.vlines(np_good_k.max(), ymin=_ymin, ymax=_ymax, color='g')
    plt.annotate(" Max lag={}".format(np_good_k.max()), (np_good_k.max(), _ymin), color='g')
    plt.title('Autocorrelation function with seasonality signals')
    plt.xlabel('Lag')
    plt.ylabel('Autocorrelation')
    plt.annotate(
      "local",
      xy=(1, -0.07), xytext=(0, 2),
      xycoords=('axes fraction'),
      textcoords='offset points',
      size=9, ha='right', va='top',   
      bbox=dict(boxstyle="round,pad=0.3", fc="lightgray", ec="lightgray", lw=1),
      )
    plt.show()
    
  if seas_as_lag:  
    return np_acfs[1:], np_conf[1:], (np.argwhere(np_season[1:]) + 1).ravel()
  else:
    return np_acfs[1:], np_conf[1:], np_season[1:]


def get_valid_lags(data, verbose=True, max_k=None, return_acf=False):
  np_acfs, np_confs, np_seas = calc_acfs(
    data=data, 
    max_k=max_k,
    )
  np_seas = np.arange(1, np_seas.shape[0]+1)[np_seas]
  if verbose:
    for _s in np_seas:
      print("  has seasonality={}".format(_s))
  if return_acf:
    return np_seas, np_acfs[np_seas]
  else:
    return np_seas


def rolling_mean(ts_init, window, center=True):
  ser = pd.Series(ts_init)
  return ser.rolling(window, center=center).mean()

def moving_averages(ts_init, window): 
  if len(ts_init) % 2 == 0:
      ts_ma = rolling_mean(ts_init, window, center=True)
      ts_ma = rolling_mean(ts_ma, 2, center=True)
      ts_ma = np.roll(ts_ma, -1)
  else:
      ts_ma = rolling_mean(ts_init, window, center=True)

  return ts_ma


def calc_trend_coefs(series, deg=1, last_steps=None):
  if len(series.shape) == 3 and series.shape[-1] ==1:
    series = series.reshape(series.shape[:-1])
  assert len(series.shape) == 2, "Series must be (nr_series, nr_steps,) but got {}".format(series.shape)
  if last_steps is not None:
    np_y = series[:,-last_steps:]
  else:
    np_y = series
  x = np.arange(np_y.shape[1])
  coefs = np.polyfit(x, np_y.T, deg=deg)  
  return coefs.T


def get_series_trends(series, last_steps=None, k=1):
  np_series = np.array(series)
  n_series = np_series.shape[0]
  nr_steps = np_series.shape[1] if last_steps is None else last_steps
  coefs = calc_trend_coefs(np_series, deg=k, last_steps=last_steps)
  np_ticks = np.arange(nr_steps) 
  np_trends = np.zeros((n_series, nr_steps))  
  for deg in range(k+1):
    np_trends += np_ticks ** (k-deg) * coefs[:,deg].reshape(-1,1)
  return np_trends

def get_trend_preds(series, nr_steps, prev_window=None, k=1):
  np_series = np.array(series)
  n_series = np_series.shape[0]
  assert len(np_series.shape) >= 2, "Series must be (N, S) or (N, S, 1)"
  if prev_window is not None:
    assert prev_window <= np_series.shape[1], "Auto-correlation window must be smaller than actual series size"
  coefs = calc_trend_coefs(np_series, deg=k, last_steps=prev_window)
  start_tick = prev_window if prev_window is not None else np_series.shape[1]
  np_ticks = np.arange(start_tick, start_tick + nr_steps)
  np_preds = np.zeros((n_series, nr_steps))  
  for deg in range(k+1):
    np_preds += np_ticks ** (k-deg) * coefs[:,deg].reshape(-1,1)
  return np_preds


def get_linear_preds(series, nr_steps, prev_window=None, k=1):
  return get_trend_preds(series, nr_steps, prev_window=prev_window, k=k)



def pd_model(X_data, value_tensor_idx, n_steps, 
             params=None, DEBUG=False, verbose=False,
             **kwargs):
  """
  Standard definition for PD models

  Parameters
  ----------
  X_data : TYPE
    DESCRIPTION.
  value_tensor_idx : TYPE
    DESCRIPTION.
  n_steps : TYPE
    DESCRIPTION.
  params : TYPE, optional
    DESCRIPTION. The default is None.
  DEBUG : TYPE, optional
    DESCRIPTION. The default is False.
  verbose : TYPE, optional
    DESCRIPTION. The default is False.
  **kwargs : TYPE
    DESCRIPTION.

  Returns
  -------
  None.

  """
  return None



def create_lm_mat_1(np_vect, np_lags, add_trend=True, dtype='float32'):
  x = np_vect.copy().reshape(-1,1)
  n_obs = x.shape[0]
  maxlag = np.max(np_lags)
  full_mat = np.zeros((maxlag + n_obs, maxlag + 1), dtype=dtype)
  for i in range(0, int(maxlag+1)):
    full_mat[maxlag-i:maxlag-i+n_obs, maxlag-i:maxlag-i+1] = x
  # delete zero-padded 
  full_mat = full_mat[maxlag:n_obs]
  x_data = full_mat[:,1:]
  y_data = full_mat[:,:1]
  # now slice only our lags
  x_data = x_data[:, np_lags - 1]
  if add_trend:
    np_trend = np.arange(maxlag+1, n_obs+1, dtype=dtype).reshape(-1,1)
    x_data = np.hstack((np_trend, x_data))
  return x_data, y_data

def create_lm_mat_2(np_vect, np_lags, add_trend=True, dtype='float32'):
  x = np_vect.copy()
  n_obs = x.shape[0]
  n_lags = len(np_lags)
  maxlag = np.max(np_lags)
  n_train = n_obs - maxlag
  full_mat = np.zeros((n_obs - maxlag, maxlag + 1), dtype=dtype)
  for i in range(n_train):
    start = i
    end = i + maxlag + 1
    full_mat[i] = x[start:end][::-1]
  x_data = full_mat[:,1:]
  if maxlag != n_lags:
    x_data = x_data[:, np_lags -1]
  y_data = full_mat[:,:1]
  if add_trend:
    np_trend = np.arange(maxlag+1, n_obs+1, dtype=dtype).reshape(-1,1)
    x_data = np.hstack((np_trend, x_data))    
  return x_data, y_data

def create_lm_mat(np_vect, np_lags, dtype='float32', add_trend=True):
  if len(np_vect.shape)>=2:
    if np_vect.shape[1] != 1:
      raise ValueError("np_vect must be (M,) or (M,1)")
    else:
      np_vect = np_vect.ravel()
    
  if np_lags.max() != len(np_lags):
    return create_lm_mat_2(
      np_vect,
      np_lags,
      add_trend,
      dtype=dtype,
      )
  else:
    return create_lm_mat_1(
      np_vect,
      np_lags,
      add_trend,
      dtype=dtype,
      )
  

STANDARD_FIRST_SALE_CONFIDENCES = {
      0.10 : {
        'interval' : [0,60], 
        'description' : 'Minimal confidence for sales that started max 60 days ago'
        },

      0.20 : {
        'interval' : [61,90],
        'description' : 'Very low confidence for sales that started max 60-90 days ago'
        },

      0.35 : {
        'interval' : [91,180],
        'description' : 'Low confidence for sales that started 90-180 days ago'
        },

      0.50 : {
        'interval' : [181,360],
        'description' : 'Average confidence for sales that started 181-360 days ago'
        },

      0.70 : {
        'interval' : [361,500],
        'description' : 'Superior confidence for sales that started 361-500 days ago'
        },

      0.90 : {
        'interval' : [500, np.inf],
        'description' : 'Max confidence for sales that started 500+ days ago'
        },

      }

STANDARD_LAST_SALE_CONFIDENCES = {
      0.10 : {
        'interval' : [180,np.inf], 
        'description' : ''
        },

      0.30 : {
        'interval' : [90,180],
        'description' : ''
        },

      0.40 : {
        'interval' : [60,90],
        'description' : ''
        },

      0.50 : {
        'interval' : [30,60],
        'description' : ''
        },

      0.70 : {
        'interval' : [10,30],
        'description' : ''
        },

      0.90 : {
        'interval' : [0, 10],
        'description' : ''
        },

      }
  
DEBUG_FIRST_SALE_CONFIDENCES = {
      0.10 : {
        'interval' : [0,5], 
        'description' : 'Minimal confidence for sales that started max 60 days ago'
        },

      0.20 : {
        'interval' : [6,9],
        'description' : 'Very low confidence for sales that started max 60-90 days ago'
        },

      0.35 : {
        'interval' : [10,15],
        'last_sale_before_today' : None,
        'description' : 'Low confidence for sales that started 90-180 days ago'
        },

      0.50 : {
        'interval' : [16,20],
        'description' : 'Average confidence for sales that started 181-360 days ago'
        },

      0.70 : {
        'interval' : [21,30],
        'description' : 'Superior confidence for sales that started 361-500 days ago'
        },

      0.90 : {
        'interval' : [31, np.inf],
        'description' : 'Max confidence for sales that started 500+ days ago'
        },

      }




###
### SERIES DATA UTILS
###


def get_random_series(nr_series=50000, nr_points=750, fake_preds=False, noise_max=None):
  from . import artificial_data

  print("\x1b[1;31m" + "WARNING: `get_random_series` is obsolete! Please use artificial_data.generate_artificial_series()" + "\x1b[0m")
  return artificial_data.get_basic_random_series(
    nr_series=nr_series,
    nr_points=nr_points,
    fake_preds=fake_preds,
    noise_max=noise_max,
    )


def get_daily_series_confidence(np_series, 
                                first_day_confidence_levels=None,
                                last_day_confidence_levels=None,
                                debug=False):
  """
  This method generates a confidence probability for a range o given timeseries
  based on the availablity of data within each series.
  

  Parameters
  ----------
  np_series : np.ndarray
    DESCRIPTION.
  fist_day_confidence_levels : dict, optional
    Confidence levels based on first day with sales in timeseries. The default is None.
  last_day_confidence_levels : dict, optional
    Confidence levels based on last day with sales in timeseries. The default is None.
  debug : bool, optional
    DESCRIPTION. The default is False.

  Returns
  -------
  np_conf_mod : np.ndarray
    (N_SERIES,) vector with confidences levels

  """
  
  assert type(np_series) == np.ndarray, "np_series must be ndarray"
  np_series = np.squeeze(np_series)
  if len(np_series.shape) == 1:
    np_series = np_series.reshape(1,-1)
  assert len(np_series.shape) == 2, "np_series must be [N_SERIES, N_HIST_STEPS]"
  
  if first_day_confidence_levels is None:
    if debug:
      first_day_confidence_levels = DEBUG_FIRST_SALE_CONFIDENCES
    else:
      first_day_confidence_levels = STANDARD_FIRST_SALE_CONFIDENCES
  
  if last_day_confidence_levels is None:
    if not debug:
      last_day_confidence_levels = STANDARD_LAST_SALE_CONFIDENCES
    
  np_non_zero = np_series != 0
  n_days = np_series.shape[1]
  n_series = np_series.shape[0]
  first_sale_before_today = n_days - np_non_zero.argmax(1)
  last_sale_before_today = np_non_zero[:,::-1].argmax(1)
  n_sales_start_end = first_sale_before_today - last_sale_before_today
  # sparsity is relative to the approx history size for each series
  sparsity = np_non_zero.sum(1) / n_sales_start_end
  first_day_confidences = np.zeros(n_series)
  last_day_confidences = np.ones(n_series)
  for ser in range(n_series):
    ser_first = first_sale_before_today[ser]
    ser_last = last_sale_before_today[ser]    
    for conf_first_day in first_day_confidence_levels:
      cnf_first = first_day_confidence_levels[conf_first_day]['interval']
      found = False
      if cnf_first[0] <= ser_first and cnf_first[1] >= ser_first:
        found = True
        break
    if found:
      first_day_confidences[ser] = conf_first_day
    # now for second day
    if last_day_confidence_levels is not None:
      for conf_last_day in last_day_confidence_levels:
        cnf_last = last_day_confidence_levels[conf_last_day]['interval']
        if cnf_last[0] <= ser_last and cnf_last[1] >= ser_last:
          found=True
          break
      if found:
        last_day_confidences[ser] = conf_last_day
        
  np_conf_mod = first_day_confidences * last_day_confidences * sparsity
  if debug:
    print("Non zero count:      {}".format(np_non_zero.sum(1)))
    print("Sarsity confidence:  {}".format(sparsity))
    print("First sale days ago: {}".format(first_sale_before_today))
    print("Confidence 1st day:  {}".format(first_day_confidences))
    print("Confidence last day: {}".format(last_day_confidences))
    print("Final confidence:    {}".format(np_conf_mod))
  return np_conf_mod



def get_past_periods_test_datasets(X, steps, freq):
  """
  generates tuple with test datasets
  """
  __RULES = {
    'M' : {
        'PERIODS' : [12, 24],
      },
    'W' : {
        'PERIODS' : [4, 52, 104],
      },
    'D' : {
        'PERIODS' : [30, 365, 730],
      },
    
    }
  TestData = namedtuple(
    'TestData', 
    ['X', 'y', 'look_back', 'freq']
    )
  freq = freq.upper()[0]
  if freq not in __RULES:
    raise ValueError("Past periods analysis for '{}' not implemented!".format(freq))
  periods = __RULES[freq]['PERIODS']
  if steps not in periods:
    periods = [steps] + periods
  result = []
  for p in periods:
    if p >= steps and X.shape[1] > p:
      start = -p
      end = (-p + steps) if steps < p else None
      np_per = X[:, start:end]
      x_data = X[:,:start]
      y_data = np_per
      # if x_data.shape[1] >= y_data.shape[1]:
      d = TestData(
        X=x_data,
        y=y_data,
        look_back=p,
        freq=freq,
        )
      result.append(d)        
  return result
  
          
        
        
  
if __name__ == '__main__':
  np.set_printoptions(
    suppress=True,
    # precision=3,
    # floatmode='fixed',
    threshold=1000,
    linewidth=1000,
    )
  #
  # test series confidence
  #
  # a = np.random.randint(0,5, size=(10,30))
  # a[0,:-4] = 0
  # a[1,:-8] = 0
  # a[2,:-15] = 0
  # a[3,::2] = 0
  # get_daily_series_confidence(
  #   np_series=a,
  #   first_day_confidence_levels=None,
  #   last_day_confidence_levels=None,
  #   debug=True,
  #   )
  #
  # test slicing
  #
  M = np.array([np.arange(0.01,0.13, step=0.01)] * 2)
  M = np.concatenate((M+1,M+2,M+3), axis=-1)[:,:-5]
  W = np.array([np.arange(0.01,0.53, step=0.01)] * 2)
  W = np.concatenate((W+1,W+2,W+3), axis=-1)[:,:-10]
  D = np.array([np.arange(0.001,0.366, step=0.001)] * 2)
  D = np.concatenate((D+1,D+2,D+3), axis=-1)[:,:-50]

  vm = get_past_periods_test_datasets(
    X=M,
    steps=12,
    freq='M'
    )

  
  vw = get_past_periods_test_datasets(
    X=W,
    steps=8,
    freq='W'
    )
  
  vd = get_past_periods_test_datasets(
    X=D,
    steps=15,
    freq='D'
    )
  
  


from naeural_core.core_logging import SBLogger
from naeural_core.local_libraries.nn.th.training.data.simple_image_classifier import SimpleImageClassifierDataLoaderFactory

if __name__ == '__main__':

  log = SBLogger()
  dev_data_factory = SimpleImageClassifierDataLoaderFactory(
    log=log,
    num_workers=0,
    batch_size=16,
    path_to_dataset='./_local_cache/_data/POSE/2_dev',
    data_subset_name='dev',
    image_height=200,
    image_width=100,
    device='cpu',
    files_extensions=None # default vision behavior (.jpg, .png etc)
  )

  dev_data_factory.create()






import cv2

from naeural_core import Logger
from decentra_vision.draw_utils import DrawUtils

if __name__ == '__main__':
  cfg_file = 'main_config.txt'
  log = Logger(
    lib_name='SB',
    config_file=cfg_file,
    max_lines=1000,
    TF_KERAS=False
  )

  painter = DrawUtils(log=log)
  img = cv2.imread('xperimental/covid1.jpg')
  img = painter.cross_from_tlbr(
    image=img,
    tlbr=[50, 50, 100, 100]
  )
  img = painter.cross(
    image=img,
    point=[150, 150],
    color=(255, 0, 0),
    thickness=2
  )
  painter.show('test', img)
  cv2.waitKey(0)
  cv2.destroyAllWindows()



import json
import numpy as np
import paho.mqtt.client as mqtt
from naeural_core import Logger
from time import sleep

DEVICE_ID = '1234'

_CONFIG_SENDER = {
  'MQTT_HOST': 'HOSTNAME',
  'MQTT_PORT': 1883,
  'MQTT_USER': 'USERNAME',
  'MQTT_PASS': 'PASSWORD',
  'MQTT_PATH': 'PATH',
  'MQTT_TOPIC': 'config1',
  'MQTT_QOS': 1
  }

def on_connect(client, userdata, flags, rc):
  print('Connected with result code {}'.format(rc))
  return

def on_disconnect(client, userdata, rc):
  # if rc != 0:
  print('Disconnected with result code {}'.format(rc))
  return

def on_message(client, userdata, message):
  print('Received message "{}" on topic {} with QoS {}'.format(message, message.topic, str(message.qos)))
  return

def on_publish(client, userdata, mid):
  print('Publish {}, {}, {}'.format(client, userdata, mid))
  return
  
def on_log(mqttc, userdata, level, string):
  print(string)
  return
  

if __name__ == '__main__':  
  cfg_file = 'main_config.txt'
  log = Logger(
    lib_name='TST', 
    config_file=cfg_file, 
    max_lines=1000, 
    TF_KERAS=False
    )
  mqttc = mqtt.Client(
    client_id=DEVICE_ID,
    clean_session=False
    )  
  
  mqttc.username_pw_set(
    username=_CONFIG_SENDER['MQTT_USER'],
    password=_CONFIG_SENDER['MQTT_PASS']
    )
  
  mqttc.on_connect = on_connect
  mqttc.on_disconnect = on_disconnect
  mqttc.on_message = on_message
  mqttc.on_publish = on_publish
  mqttc.on_log = on_log
  
  mqttc.connect(
    host=_CONFIG_SENDER['MQTT_HOST'],
    port=_CONFIG_SENDER['MQTT_PORT']
    )

  i = 0
  topic_name = '{}/{}'.format(_CONFIG_SENDER['MQTT_PATH'], _CONFIG_SENDER['MQTT_TOPIC'])
  # while True:
  for i in range(100):
    img = np.random.randint(0, 256, (2000, 3000, 3), dtype=np.uint8)
    img_base64 = log.np_image_to_base64(img)
    
    payload = {
      'ID': i + 1,
      'IMG': img_base64
      }
    
    result = mqttc.publish(
      topic=topic_name, 
      payload=json.dumps(payload), 
      qos=_CONFIG_SENDER['MQTT_QOS']
      )
    
    if result.rc == mqtt.MQTT_ERR_QUEUE_SIZE:
      raise ValueError('Message is not queued due to ERR_QUEUE_SIZE')
    
    try:
      while not result.is_published():
        mqttc.loop(timeout=0.01)
        # result.wait_for_publish(0.01)
    except:
      print('Exception')
    sleep(0.1)
    
    





















"""
INFO:
  aspects in any config_data using class including DecentrAIObject !!!
  as a rule-of-the-thumb this mixin should be used as:
    - add_config_data(dct_config) - this will append/update new config to the config_data
    - create_config_handlers - after we have final config_data we can create implicit `cfg_` handlers
    - validation  & run_validation_rules - finally check if all is ok

TODO:
  
  
  
"""
from copy import deepcopy
from collections import deque
import traceback

class VALIDATION_KEYS:
  TYPE = 'TYPE'
  MIN_VAL = 'MIN_VAL'
  MAX_VAL = 'MAX_VAL'
  MIN_LEN = 'MIN_LEN'
  EXCLUDED_LIST = 'EXCLUDED_LIST'
  ACCEPTED_VALUES = 'ACCEPTED_VALUES'
  
def get_validation_keys():
  res = []
  for k in VALIDATION_KEYS.__dict__:
    if k.isupper():
      res.append(VALIDATION_KEYS.__dict__[k])
  return res

class CONST(VALIDATION_KEYS):
  RULES = 'VALIDATION_RULES'
  FUNC = '_cfg_validate_'
  PROP_PREFIX = 'cfg_'
  

from functools import partial
def getter(slf, key=None):
  val = slf.config_data.get(key)
  if isinstance(val, dict):
    return val.copy()
  else:
    return val

def get_all_elements_from_deque(deq):
  lst = []
  while len(deq) > 0:
    x = deq.popleft()
    lst.append(x)
  return lst
#enddef

class _ConfigHandlerMixin(object):

  def __init__(self):
    self._valid_errors = deque(maxlen=100)
    self._valid_warnings = deque(maxlen=100)
    self._valid_infos = deque(maxlen=100)
    self.__cfg_ready = False
    super(_ConfigHandlerMixin, self).__init__()
    return
  
  @property
  def ready_cfg_handlers(self):
    return self.__cfg_ready
  
  
  def needs_update(self, dct_newdict, except_keys):
    """Check if we need to perform a config update: if a new dict is different from current config_data

    Parameters
    ----------
    dct_newdict : dict
        The new dict to be checked
    except_keys : list
        list of keys to be excluded from check

    Returns
    -------
    bool, list
        need to update or not and list of keys that are different
    """
    result = False
    updates = []
    if isinstance(dct_newdict, dict):      
      for k in dct_newdict:
        if self.config_data.get(k) != dct_newdict[k]:
          updates.append(k)
        if (k not in except_keys) and ((k not in self.config_data) or (self.config_data.get(k) != dct_newdict[k])):
            result = True            
    return result, updates
  

  def _merge_prepare_config(
    self, 
    default_config=None, 
    delta_config=None, 
    uppercase_keys=True, 
    verbose=0,
    debug=False
  ):
    if default_config is None:
      default_config = vars(self).get('_default_config')
      msg_src = '_default_config '
    else:
      msg_src = ''

    if delta_config is None:
      delta_config = vars(self).get('_upstream_config')
      if delta_config is None:
        return default_config
      msg_delta = "using `_upstream_config` ({} keys)".format(len(delta_config))
    else:
      msg_delta = "using new cfg ({} keys)".format(len(delta_config))

    if debug:
      msg_delta += ": {}".format(list(delta_config.keys()))
    
    # start delta_config prepare
    # now deepcopy delta_config to a local variable to ensure mutable values will not
    # be modified in _upstream_config thus triggering a diff from actual config    
    delta_config = deepcopy(delta_config)
    # convert to upper default    
    if uppercase_keys:
      delta_config = {k.upper():v for k,v in delta_config.items()}
    # end delta_config prepare 
    lst_msg = []
    
    lst_msg.append("Updating {} {}config {}:".format(
      self.__class__.__name__, msg_src,
      msg_delta
      )
    )
    if debug and hasattr(self, 'log'):
      str_init = '\n{}'.format(self.log.dict_pretty_format(default_config))
      str_delta = '\n{}'.format(self.log.dict_pretty_format(delta_config))
      self.P("Update info:\n************ Initial:\n{}\n************ Modifications:\n{}".format(str_init, str_delta))
      
    if default_config is None:
      self.P("WARNING: no default config was provided at {} startup!".format(
        self.__class__.__name__), color='r')
      final_config = {}
    else:
      final_config = deepcopy(default_config)
    #endif
    all_keys = set(final_config.keys()).union(set(delta_config.keys()))
    for k in all_keys:
      if final_config.get(k) == delta_config.get(k) or k not in delta_config:
        if debug:
          lst_msg.append("  {}={}".format(k, final_config.get(k)))
      else:
        if k not in final_config:
          if verbose > 1:
            lst_msg.append("  {}={} [NEW]".format(k, delta_config[k]))
          else:
            lst_msg[-1] += k + '; '
          #endif verbose
        elif final_config[k] != delta_config[k]:
          if verbose > 1:
            lst_msg.append("  {}={} -> {}={}".format(
              k, final_config[k], k, delta_config[k])
            )
          else:
            lst_msg[-1] += k + '; '
          #endif verbose
        final_config[k] = delta_config[k]
        #endif new key or value
      #endif key not in delta_config
    #endfor all keys
    if len(lst_msg) > 0 and verbose > 0:      
      full_dump = "\n".join(["    " + x for x in lst_msg])        
      self.P(full_dump.lstrip(), color='b')
    return final_config
  
  def update_config_data(self, new_config_data):
    return self.add_config_data(new_config_data=new_config_data)

  def add_config_data(self, new_config_data):
    lst_msg = []
    if not isinstance(new_config_data, dict):
      self.P("`config_data` must be updated with a dict. Received '{}'".format(type(new_config_data)))
      return False
    
    lst_msg.append("Updating '{}' configuration...".format(self.__class__.__name__))
    
    if not hasattr(self, 'config_data'):
      self.config_data = {}

    all_keys = set(self.config_data.keys()).union(set(new_config_data.keys()))
    for k in all_keys:
      if (self.config_data.get(k) == new_config_data.get(k)) or (k not in new_config_data):
        lst_msg.append("  {}={}".format(k, self.config_data[k]))
      else:
        if k not in self.config_data:
          lst_msg.append("  {}={} [NEW]".format(k, new_config_data[k]))
        elif self.config_data[k] != new_config_data[k]:
          lst_msg.append("  {}={} -> {}={}".format(
            k, self.config_data[k], k, new_config_data[k]
            )
          )
        self.config_data[k] = new_config_data[k]
    if len(lst_msg) > 0:
      full_dump = "\n".join(["    " + x for x in lst_msg])        
      self.P(full_dump.lstrip(), color='b')
    return True

  
  def create_config_handlers(self, verbose=0):    
    if hasattr(self, 'config_data') and isinstance(self.config_data, dict) and len(self.config_data) > 0:
      res = []
      for k in self.config_data:
        func_name = self._get_prop(k)
        if not hasattr(self, func_name):
          # below is a bit tricky: using a lambda generates a non-deterministic abnormal behavior
          # the ideea is to create a global func instance that wil be then loaded on the class (not instance)
          # as a descriptor object - ie a `property`. "many Bothans died to bring the plans of the Death Star..."
          fnc = partial(getter, key=k) # create the func
          cls = type(self) # get the class
          fnc_prop = property(fget=fnc, doc="Get '{}' from config_data".format(k)) # create prop from func 
          setattr(cls, func_name, fnc_prop) # set the prop of the class
          res.append(func_name)
      if len(res) > 0 and verbose > 1:
        self.P("Created '{}' config_data handlers: {}".format(self.__class__.__name__, res), color='b')
      self.__cfg_ready = True
    return
  
  
  def _cfg_check_type(self, cfg_key, types):
    val = self._get_val_by_prop(cfg_key)
    if not isinstance(val, types):
      msg = "`{}` config key '{}={}' requires type `{}` - found `{}`".format(
        self.__class__.__name__, 
        cfg_key, val, [x.__name__ for x in types] if isinstance(types, tuple) else types.__name__,
        type(val).__name__
      )
      return False, msg
    return True, ''
  
  def _cfg_check_min_max(self, cfg_key, dct_rules):
    res = True
    msg = None
    _min = dct_rules.get(CONST.MIN_VAL)
    _max = dct_rules.get(CONST.MAX_VAL)
    val = self._get_val_by_prop(cfg_key)
    if _min is not None and val < _min:
      msg = "`{}` config key '{}={}' of type `{}` requires value > {}".format(
        self.__class__.__name__, cfg_key, val, dct_rules.get(CONST.TYPE), _min,
      )
      return False, msg
    
    if _max is not None and val > _max:
      msg = "`{}` config key '{}={}' of type `{}` requires value < {}".format(
        self.__class__.__name__, cfg_key, val, dct_rules.get(CONST.TYPE),  _max,
      )
      return False, msg
    return res, msg

  def _cfg_check_exclusions(self, cfg_key, dct_rules):
    _excl_lst = dct_rules.get(CONST.EXCLUDED_LIST)
    val = self._get_val_by_prop(cfg_key)
    if _excl_lst is not None and val in _excl_lst:
      msg = "`{}`  key '{}={}' in exclutions {}".format(
        self.__class__.__name__, cfg_key, val, _excl_lst,
      )
      return False, msg
    return True, ''


  def _cfg_check_accepted(self, cfg_key, dct_rules):
    _accept_lst = dct_rules.get(CONST.ACCEPTED_VALUES)
    val = self._get_val_by_prop(cfg_key)
    if _accept_lst is not None and val not in _accept_lst:
      msg = "`{}`  key '{}={}' not in accepted values {}".format(
        self.__class__.__name__, cfg_key, val, _accept_lst,
      )
      return False, msg
    return True, ''
  
  def _cfg_validate_int(self, cfg_key, dct_rules):
    is_ok, msg = self._cfg_check_type(cfg_key=cfg_key, types=(int,))
    if is_ok:
      is_ok, msg = self._cfg_check_min_max(cfg_key=cfg_key, dct_rules=dct_rules)
    return is_ok, msg
  
  def _cfg_validate_float(self, cfg_key, dct_rules):    
    is_ok, msg = self._cfg_check_type(cfg_key=cfg_key, types=(int,float))
    if is_ok:
      is_ok, msg = self._cfg_check_min_max(cfg_key=cfg_key, dct_rules=dct_rules)
    return is_ok, msg
  
  
  def _cfg_validate_str(self, cfg_key, dct_rules):
    res1, msg1 = self._cfg_check_type(cfg_key=cfg_key, types=(str,))
    res, msg = res1, msg1
    if res1:
      res2, msg2 = self._cfg_check_exclusions(cfg_key=cfg_key, dct_rules=dct_rules)
      res3, msg3 = self._cfg_check_accepted(cfg_key=cfg_key, dct_rules=dct_rules)
      val = self._get_val_by_prop(cfg_key)
      _min_len = dct_rules.get(CONST.MIN_LEN, 0)
      res4 = True
      msg4 = None
      if res1: # must have valid type to check next
        if _min_len is not None and len(val) < _min_len:
          msg4 = "`{}` config key '{}={}' of type `{}` must have at least {} chars".format(
            self.__class__.__name__, cfg_key, val, dct_rules.get(CONST.TYPE), _min_len,
          )
          res4 = False
      res = res1 and res2 and res3 and res4
      msg = msg1 if not res1 else (msg2 if not res2 else (msg3 if not res3 else msg4))
    return res, msg      

  def _cfg_validate_list(self, cfg_key, dct_rules):
    res1, msg1 = self._cfg_check_type(cfg_key=cfg_key, types=(list,))
    res2, msg2 = self._cfg_check_exclusions(cfg_key=cfg_key, dct_rules=dct_rules)
    res3 = True
    msg3 = None
    if res1: # must have valid type to check next
      val = self._get_val_by_prop(cfg_key)
      _min_len = dct_rules.get(CONST.MIN_LEN, 0)
      if _min_len is not None and len(val) < _min_len:
        msg3 = "`{}` config key '{}={}' of type `{}` must have at least {} elements, found {}".format(
          self.__class__.__name__, cfg_key, val, dct_rules.get(CONST.TYPE), _min_len, len(val)
        )
        res3 = False
    res = res1 and res2 and res3
    msg = msg1 if not res1 else (msg2 if not res2 else msg3)
    return res, msg      
  
    
  def _get_prop(self, k):
    return CONST.PROP_PREFIX + k.lower().replace('#','_')
  
  def _get_val_by_prop(self, k):
    f = getattr(self, self._get_prop(k), None)
    val = f
    return val
    
  
  def run_validation_rules(self, verbose=0, debug_verbose=False):
    result_state = {'RESULT': True, 'MSGS' : []}
    def __fail(s):
      result_state['RESULT'] = False
      result_state['MSGS'].append(s)
      if debug_verbose:
        self.P("  " + s, color='r')
      return
    if verbose:
      self.P("Validating configuration for '{}'...".format(self.__class__.__name__), color='b')
    if hasattr(self, 'config_data') and isinstance(self.config_data, dict):
      dct_validation = self.config_data.get(CONST.RULES, {})
      for k in dct_validation:
        # run each config key present in validation area
        prop_getter = self._get_prop(k)
        has_prop_getter = prop_getter in dir(self) #getattr(self, prop_getter, None) # get the getter
        if not has_prop_getter:
          __fail("No config getter `{}` for VALIDATION key '{}' - please run `create_config_handlers` before validation or check maybe there is no CONFIG key defined".format(
            prop_getter, k)
          )
          continue
        if k not in self.config_data: 
          # have getter but no key in json - bad programming?
          self.P("  Key '{}' found in validation not present in `config_data` however `{}` has getter `{}`. This is not a accepted approach, please define '{}' key-val in `_CONFIG_DATA` ".format(
            k, self.__class__.__name__, prop_getter, k.upper(), ), color='r'
          )
        dct_rules = dct_validation.get(k, {})
        if len(dct_rules) > 0:
          # now we get the actual type of the value 
          configured_type = dct_rules.get(CONST.TYPE)
          _type = None
          if isinstance(configured_type, str) and len(configured_type) > 1:
            try:
              _type = eval(configured_type)
              if type(_type) != type:
                _type = None                
            except:
              _type = None
            if _type is None:
              __fail("TYPE '{}' of '{}' pre-validation failed - unknown type!".format(configured_type, k))
              continue
          else:
            # not type info maybe not big error
            if verbose > 0:
              if configured_type is None:
                self.P("  No type information found for '{}': {}".format(k, configured_type), color='r')
              else:
                self.P("  Complex type information found for '{}': {}".format(k, configured_type))
          # endif type extraction
          if _type in [int, float, str, list]:            
            # create validation function out of this mixin available funcs
            str_func = CONST.FUNC + _type.__name__
            func = getattr(self, str_func, None)
            res = True # assume good
            if func is None:
              # if we use a predefined type then we must have the validation
              self.P("  No default handler for type '{}' config key validation".format(_type.__name__), color='r')
              # maybe we can put res = False here?
            else:
              msg = ''
              # now we run the validation function
              res = func(k, dct_rules)
              if isinstance(res, tuple):
                res, msg = res
            # end run-or-fail validation
            if not res:
              # validation failed
              __fail("Config validation for '{}={}' of '{}' failed: {}".format(
                k, getattr(self, prop_getter, None), self.__class__.__name__, msg)
              )
              # here we can break for but we leave to see what other error we have   
            elif verbose:
              self.P("  Config validation for key '{}' succeeded".format(k), color='b')
            #endif failed or not
              
          # end of known types
          else:
            # if other type than known ones
            if verbose > 1:
              self.P("  Unavailable basic handler for type '{}' for config key '{}' - running auto check".format(
                configured_type, k,
                ), 
                color='y'
              )
            # but we can still run the type check without special handler
            # no need to check if _type is not None         
            if _type is None and isinstance(configured_type, list):
              _type = tuple([eval(x) for x in configured_type])
            res, msg = self._cfg_check_type(cfg_key=k, types=_type)
            if not res:
              __fail("Automatic checking of unhandled type `{} <{}>`: {}".format(
                configured_type,
                _type,
                msg
                )
              )
        else:
          self.P("  Empty rules info for '{}'".format(k), color='y')
        # end if rules parsing
      # end for each key validation 
      if result_state['RESULT']:
        self.P("Automatic validation for instance of `{}` is successful".format(
          self.__class__.__name__), 
          color='g'
        )
      else:
        __fail('Automatic validation for instance of `{}` FAILED.'.format(self.__class__.__name__))
    # end if we have config and validation date
    else:
      __fail("No `config_data` for '{}'".format(self.__class__.__name__))
    
    result = result_state['RESULT']
    fail_msgs = result_state['MSGS']
    return result, fail_msgs

  def get_errors(self):
    return get_all_elements_from_deque(self._valid_errors)

  def get_warnings(self):
    return get_all_elements_from_deque(self._valid_warnings)

  def get_infos(self):
    return get_all_elements_from_deque(self._valid_infos)

  def add_error(self, msg):
    self._valid_errors.append(msg)
    return

  def add_warning(self, msg):
    self._valid_warnings.append(msg)
    return

  def add_info(self, msg):
    self._valid_infos.append(msg)
    return

  def validate(self, raise_if_error=True, verbose=0):
    _, fail_msgs = self.run_validation_rules(verbose=verbose)

    for msg in fail_msgs:
      self.add_error(msg)

    for method_name, func in self.log.get_class_methods(self.__class__, include_parent=True):
      if method_name.startswith('validate_'):
        try:
          func(self)
        except:
          self.add_error("Programming error in validation method '{}'.\n{}".format(method_name, traceback.format_exc()))

    lst_errors = self.get_errors()
    lst_warnings = self.get_warnings()
    lst_infos = self.get_infos()

    has_errors = len(lst_errors) > 0
    is_valid = not has_errors

    for x in lst_infos:
      self.P(x)

    for x in lst_warnings:
      self.P(x, color='y')

    for x in lst_errors:
      self.P(x, color='r')

    if raise_if_error and has_errors:
      raise ValueError("Errors occured while validating: {}".format(lst_errors))

    return is_valid
  
  def setup_config_and_validate(self, dct_config, verbose=0):
    if verbose:
      self.P("Resetting config_data ...")
    self.config_data = dct_config
    if verbose:
      self.P("Running config handler creation...")
    self.create_config_handlers(verbose=verbose)
    if verbose:
      self.P("Running validation...")
    self.validate(verbose=verbose)
    return




from naeural_core import Logger
if __name__ == '__main__':
  l = Logger("PLG", base_folder='.', app_folder='_local_cache')
  paths = l.get_all_subfolders("naeural_core/xperimental/numpy", as_package=True)
  for path in paths:
    l.P(path)


# -*- coding: utf-8 -*-
"""
  
  
TODO:
  - on every error deallocate parallel serving process
  - use faster & non blocking communication  
  - Raise warning when preprocessing on a large amount of images !
  
  - model cloning for same definition but DIFFERENT weights

windows: set PYTHONPATH=%PYTHONPATH%;C:\work\ 
  


"""

import os
import threading
import traceback


from time import sleep, time, strftime, gmtime
from collections import OrderedDict, defaultdict

# os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

from naeural_core import constants as ct
from naeural_core.utils.system_shared_memory import NumpySharedMemory, replace_ndarray_with_shm, replace_shm_with_ndarray

from naeural_core.manager import Manager
from naeural_core.serving.base.serving_utils import comm_generator

_FULL_DEBUG = True

class SMConst:
  PCOUNT = 'PCOUNT'
  PTIME = 'PTIME'
  CREATED_DATE = 'CREATED_DATE'
  DATA = 'DATA'
  TYPE = 'TYPE'
  RESULT = 'RESULT'
  TIMERS = 'TIMERS'
  ABNORMAL_EXIT = 'ABNORMAL_EXIT'
  PREDICT = 'PREDICT'
  READY = 'READY'
  GET_CONFIG = 'GET_CONFIG'
  STOP = 'STOP'
  SERVER_NAME = 'SERVER_NAME'
  
  COMM = 'COMM'
  SERVER = 'SERVER'
  PID = 'PID'
  COVERED_BY = 'COVERED_BY'
  SHM = 'SHM'
  DEFAULT_COMM_METHOD = 'shm'
  
  CMD_ID = 'CMD_ID'
  TIMESTAMP = 'TIMESTAMP'
  
  BASE_PROPS = [COMM, SERVER, PID]
  
  SERVING_ENVIRONMENT = 'SERVING_ENVIRONMENT'
  MAX_WAIT_TIME_DEFAULT = 5
  MAX_WAIT_TIME_KEY = 'MAX_WAIT_TIME'
  
  MAX_WAIT_TIME_MULTIPLIER_DEFAULT = max(2, 100 // MAX_WAIT_TIME_DEFAULT) # this is used only for startups
  
  MAX_INPROCESS_TIME = 12000
  
  SERVING_MAX_TRIES = "SERVING_MAX_TRIES"
  SERVING_MAX_TRIES_DEFAULT = 5
  
  
def get_raw_server(log, server_name):
  """
  Simple utility function that enables quick experimentation of serving process without Serving Manager

  Parameters
  ----------
  log : Logger

  server_name : str


  """
  from naeural_core.manager import Manager
  mgr = Manager(log=log)
  _, class_name, _Class, _default_config = mgr._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_SERVING_PLUGINS,
      name=server_name,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_SERVING_PLUGINS,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_SERVING_PLUGINS,
      safe_imports=ct.PLUGIN_SEARCH.SERVING_SAFE_IMPORTS,
      safety_check=True,
    )  
  
  _server = _Class(
      server_name=server_name,
      comm_eng=None,
      inprocess=True,
      default_config=_default_config,
      log=log,
    )  
  _server.predict_server_name = server_name.upper()
  return _server
    
  

class ServingManager(Manager):
  def __init__(self,
               shmem,
               server_names=None,
               owner=None,
               full_debug=None,
               log_timeouts_period=3600,
               monitor_each_k_inferences=10,
               monitor_callback=None,
               **kwargs):
    """
    The main ServingManager

    Parameters
    ----------
    server_names : list[str]
      the initial list of servers (better just empty list).
      
    full_debug : TYPE, optional
      show "extreme" debug info. The default is False.
            
      
    **kwargs : TYPE
      must pass Logger instance.


    Usage/tools:
      
      `start_server(server)`                      : will start a serving process based on the given name
        
      `stop_server(name)`                         : gracefully stop the server
      
      `stop_all_servers()`                        : gracefully stop all servers
      
      `get_server_config(name, key)`              : returns full serving process config or just a key
      
      `predict(name, inputs)`                     : run prediction using inputs (can be anything)
      
      `predict_parallel(dct)`                     : run all given {server:input} pairs in parallel
      
      `get_predict_time(name)`                    : get mean timing for the given serving process - depends on
                                                    individual calls and individual batch sizes
      
      `get_predict_parallel_time()`               : returns mean parallel predict time depending of all jobs 
                                                    that have been given
                                          
      `keep_active_servers(active_server_names)`  : will keep running only the requested server names.
                                                    will close any started server that is not found in active_server_names
                                                    will start any un-started server that is found in active_server_names

    """
    self.shmem = shmem
    # TODO: maybe add SharedMemoryManager to each serving process instance under category=ct.SHMEM.SERVING
    # however keep in mind that it will not work for simple dicts!!!
      
    if server_names is None:
      server_names = []

    if full_debug is None:
      full_debug = False

    self.owner = owner
      
    self._monitor_each_k_inferences = monitor_each_k_inferences
    self._monitor_counter = 0
    self._monitor_callback = monitor_callback

    self.full_debug = full_debug
    self.log_timeouts_last_ts = None
    self.log_timeouts_period = log_timeouts_period
    self._servers = {}
    self._timeouts = {}
    self._collector_call_counter = 0
    self.__cmd_count = 0
    self.__server_failures = defaultdict(lambda : 0)
    self._server_names = server_names
    self.inprocess_predict = None
    self._main_tid = threading.get_ident()
    prefix_log = kwargs.pop('prefix_log','[SMGR]')
    super(ServingManager, self).__init__(prefix_log=prefix_log, **kwargs)
    return
  
  def P(self, s, color=None, **kwargs):
    s = "" + s # maybe think a extra prefix
    super().P(s, color=color, **kwargs)
    return
  
  def get_server_name(self, server_name):
    """
    Returns the stringified server name in uppercase
    Args:
      server_name: str or list[str] or tuple[str] - the server name(s)
    Returns:
      res : str - the stringified server name in uppercase
    """
    str_id = None
    if isinstance(server_name, str):
      str_id = server_name.upper()
    elif isinstance(server_name, (list, tuple)):
      str_id = '_'.join([x.upper() for x in server_name])
    else:
      self.P("Unknown server name format {}".format(server_name), color='error')
    return str_id
  
  def startup(self):
    super().startup()
    self.P("Initialising {} with following environment:\n{}".format(
      self.__class__.__name__,
      self.log.dict_pretty_format(self.cfg_serving_environment)
    ), color='b')
    self._create_servers()
    return

  @property
  def serving_pids(self):
    pids = []
    for s, dct_s in self._servers.items():
      pids.append(dct_s[SMConst.PID])
    return list(set(pids))

  @property
  def total_timeouts(self):
    return sum(self._timeouts.values()) if len(self._timeouts.keys()) > 0 else 0

  @property
  def cfg_serving_environment_dict(self):
    return self.log.config_data.get(SMConst.SERVING_ENVIRONMENT, {})
  
  @property
  def cfg_serving_environment(self):
    return self.cfg_serving_environment_dict

  @property
  def default_cuda(self):
    configured_cuda = self.cfg_serving_environment.get(ct.DEFAULT_CUDA, 'cuda:0')
    gpu_info = self.log.gpu_info()
    if gpu_info is None or len(gpu_info) == 0:
      return 'cpu'
    return configured_cuda
  
  @property
  def _server_collector_timedelta(self):
    return self.cfg_serving_environment.get(ct.SERVING.SERVER_COLLECTOR_TIMEDELTA, 3600)
    
  @property
  def _server_collector_delay(self):
    return self.cfg_serving_environment.get(ct.SERVING.SERVER_COLLECTOR_DELAY, 5)


  @property
  def must_check_blocked_inprocess_serving(self):
    return self.cfg_serving_environment.get(ct.SERVING.CHECK_BLOCKED_INPROCESS_SERVING, True)
  
  @property
  def max_wait_time_multiplier(self):
    return self.cfg_serving_environment.get(ct.SERVING.MAX_WAIT_TIME_MULTIPLIER, SMConst.MAX_WAIT_TIME_MULTIPLIER_DEFAULT)
  
  ###
  ### PRIVATE/PROTECTED
  ###

  def get_timeouts(self, server_name):
    if server_name not in self._timeouts:
      self._timeouts[server_name] = 0
    return self._timeouts[server_name]

  def increment_timeouts(self, server_name):
    if server_name not in self._timeouts:
      self._timeouts[server_name] = 0
    self._timeouts[server_name] += 1
    return self._timeouts[server_name]

  def _assign_server(self, replaced_server, main_server):
    assert main_server in self._servers
    if replaced_server not in self._servers:
      self._servers[replaced_server] = {}
    for prop in SMConst.BASE_PROPS:
      self._servers[replaced_server][prop] = self._servers[main_server][prop]
    self._servers[replaced_server][SMConst.COVERED_BY] = main_server
    self._reset_usage(replaced_server)
    return
  
  
  def _check_equal_servers(self, replaced_server, main_server):
    result = True
    for prop in SMConst.BASE_PROPS:
      if self._servers[replaced_server][prop] != self._servers[main_server][prop]:
        result = False
    return result
  
  
  def _maybe_complete_servers(self, server_name, similar_servers, covered_servers, force_stop_alias=True): 
    server_name = self.get_server_name(server_name)
    similar_servers = [self.get_server_name(x) for x in similar_servers]
    covered_servers = [self.get_server_name(x) for x in covered_servers]
    if server_name not in self._servers:
      return
    self.P("Model '{}' parents & coverage: ".format(server_name), color='y')
    for similar_server in similar_servers:  
      if similar_server in covered_servers:
        # this similar server has been declared in covered servers
        if similar_server not in self._servers:
          # not loaded then just alias
          self.P("  Parent '{}' => will be automatically served due to COVERED_SERVERS".format(similar_server), color='m')
          self._assign_server(replaced_server=similar_server, main_server=server_name)
        else:
          # seems already loaded
          if not self._check_equal_servers(replaced_server=similar_server, main_server=server_name):
            # check if already loaded is not actually the main server (probably assigned elsewhere)
            # then (usually) stop and alias
            self.P("  WARNING: '{}' covers '{}' although '{}' is already running, please verify this behavior".format(
              server_name, similar_server, similar_server), color='r')
            if force_stop_alias and (self._servers[similar_server][SMConst.SERVER].server_name in similar_servers):
              self.stop_server(similar_server)
              self._assign_server(replaced_server=similar_server, main_server=server_name)
              # now check if no other aliases of the similar should be re-assigned to the current server
              for svr in self._servers:
                if self._servers[svr].get(SMConst.COVERED_BY) == similar_server:
                  self.P("Found '{}' already covered by '{}'. Covering '{}' by '{}'".format(
                    svr, similar_server, svr, server_name,
                    ), color='m')
                  self._assign_server(replaced_server=svr, main_server=server_name)
          else:
            self.P("  Server '{}' already covered by '{}'".format(similar_server, server_name), color='m')
        # endif similar_server in self._servers

      else:
        self.P("  Parent '{}' not in COVERED_SERVERS. No aliast added.".format(similar_server), color='y')

    for covered_server in covered_servers:
      if covered_server not in self._servers:
        # this is covered and not loaded - just make an alias
        self.P("  Model '{}' not in '{}' parents will be covered by '{}'".format(
          covered_server, server_name, server_name
          ), color='m')
        self._assign_server(replaced_server=covered_server, main_server=server_name) 
      elif covered_server not in similar_servers: 
        # this is covered and isn't a parent
        if not self._check_equal_servers(replaced_server=covered_server, main_server=server_name):
          # if main server is NOT already same as the covered server then stop & alias
          self.P("  WARNING: '{}' covers '{}' although '{}' is already running, please verify this behavior!".format(
            server_name, covered_server, covered_server), color='r')
          if force_stop_alias and (self._servers[covered_server][SMConst.SERVER].server_name in covered_servers):
            self.stop_server(covered_server)
            self._assign_server(replaced_server=covered_server, main_server=server_name)        
        else:
          self.P("  Server '{}' already covered by '{}'".format(covered_server, server_name), color='m')          
    return
  
  def _download_and_load_config(self, url, server_name, force_download=True):
    fn_custom_model_definition = server_name + '_def.txt'
    self.log.maybe_download_model(
      url, fn_custom_model_definition, force_download=force_download
    )
    dct_config = self.log.load_models_json(fn_custom_model_definition)
    return dct_config

  def get_comm_method(self):
    """
    Get the communication method used by the serving processes.
    Options:
    - 'default': will send messages from manager to serving processes without
                storing the ndarrays in shared memory
    - 'shm': will send messages from manager to serving processes in the same manner,
            but will replace any ndarrays from the message with references to shared memory
    Returns
    -------

    """
    return self.cfg_serving_environment.get(ct.SERVING.COMM_METHOD, SMConst.DEFAULT_COMM_METHOD)

  def _server_wait_time(self, server_name):
    server_name = self.get_server_name(server_name)
    if server_name not in self._servers:
      return SMConst.MAX_WAIT_TIME_DEFAULT
    _server = self._servers[server_name][SMConst.SERVER]
    # TODO: move this to received config 
    wait_time = _server.CONFIG[SMConst.MAX_WAIT_TIME_KEY]
    return wait_time

  def _create_server(self, server_name, server_class_name=None, upstream_config=None, inprocess=False):
    upstream_config = upstream_config or {}
    if server_name in self._servers:
      self.P("Serving process '{}' already created".format(server_name), color='r')
      return server_name
    if server_class_name is None:
      server_class_name = server_name
    self.P("Creating {}serving process '{}'({})...".format(
      "INPROCESS " if inprocess else "PARALLEL ",
      server_name, server_class_name))
    self.log.start_timer('create_server')
    _module, class_name, _Class, _default_config = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_SERVING_PLUGINS,
      name=server_class_name,
      safety_check=True,
      suffix=ct.PLUGIN_SEARCH.SUFFIX_SERVING_PLUGINS,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_SERVING_PLUGINS,     
      safe_imports=ct.PLUGIN_SEARCH.SERVING_SAFE_IMPORTS,
    )
    if _Class is None:
      self.P("Failed to load model server '{}'({}). Please define a specific class in inference/serving_plugins!".format(
        server_name, server_class_name), color='error')
      self.log.stop_timer('create_server')
      return None
    
    self.P("Loaded model serving class '{}' {}".format(class_name, _Class), color='g')

    # maybe create inter-process communication
    if not inprocess:
      comm_type = self.cfg_serving_environment.get('COMM_ENGINE', 'queue')
      comm_server, comm_client = comm_generator(engine=comm_type)
      log = None
    else:
      comm_server, comm_client = None, None
      log = self.log
    #endif
    
    # before creating the process check if the serving process is a CUSTOM_DOWNLOADABLE_MODEL
    if isinstance(_default_config, dict) and _default_config.get(ct.CUSTOM_DOWNLOADABLE_MODEL, False):
      # now load the config from the server and add it to 
      config_url = upstream_config.get(ct.CUSTOM_DOWNLOADABLE_MODEL_URL, None)
      if config_url is None:
        self.P("CUSTOM_DOWNLOADABLE_MODEL '{}' is missing CUSTOM_DOWNLOADABLE_MODEL_URL from its upstream config".format(
          server_name), color='error')
        self.log.stop_timer('create_server')
        return None
      max_tries = 3
      sleep_period = 5
      download_success = False
      current_exception = None
      dct_custom_config = None
      for try_idx in range(max_tries):
        try:
          dct_custom_config = self._download_and_load_config(config_url, server_name)
          download_success = dct_custom_config is not None
        except Exception as e:
          current_exception = traceback.format_exc()
        # endtry-except
        if not download_success:
          err_msg = (
            f"Failed to load CUSTOM_DOWNLOADABLE_MODEL '{server_name}' from URL '{config_url}'. "
            f"\nMaybe check URL. Attempt {try_idx + 1} of {max_tries}."
          )
          if current_exception is not None:
            err_msg += f"\nException: {current_exception}"
          if try_idx < max_tries - 1:
            err_msg += f"\nRetrying in {sleep_period} seconds..."
          self.P(err_msg, color='r')
          sleep(sleep_period)
        else:
          download_success = True
          break
      # endfor try_idx
      if not download_success:
        self.log.stop_timer('create_server')
        return None
      # endif download_success

      # now merge the dicts
      for k,v in dct_custom_config.items():
        _default_config[k] = v
    # end CUSTOM_DOWNLOADABLE_MODEL

    # create serving process
    _version = "0.0"
    if isinstance(_default_config, dict):
      _version = _default_config.get("MODULE_VERSION", "0.0.0")
    self.log.start_timer('create_server_new()')
    max_img_shape = self.cfg_serving_environment.get('SHM_MAX_IMAGE_SHAPE', ct.SERVING.SHM_IMG_MAX_SHAPE)
    comm_method = self.get_comm_method()
    if comm_method == 'pipe':
      self.P(f"Attempting to use pipe communication for server '{server_name}'")
    npy_shm_kwargs = {
      'mem_name': self.get_server_name(server_name),
      'mem_size': 0,  # irrelevant for buffer mode
      'np_shape': max_img_shape,
      'np_type': 'uint8',
      'is_buffer': True,
      'maxlen': self.cfg_serving_environment.get('SHM_MAX_LEN', ct.SERVING.SHM_MAX_LEN),
    }
    self.owner.set_loop_stage('3.serving.start.{}.shm_starting'.format(server_name))
    self.log.start_timer('create_server_shm_start()')
    npy_shm = None if comm_method == 'pipe' else NumpySharedMemory(**npy_shm_kwargs, create=True, log=self.log)
    if npy_shm is not None and not npy_shm.initialized:
      self.P(
        f"Failed to initialize shared memory for server '{server_name}'. Defaulting to pipe communication",
        color='error'
      )
      comm_method = 'pipe'
      npy_shm = None
    # endif initializing shared memory failed
    self.log.stop_timer('create_server_shm_start()')
    _server = _Class(
      server_name=server_name,
      comm_eng=comm_client,  # None if inprocess
      full_debug=self.full_debug,
      inprocess=inprocess,
      default_config=_default_config,
      upstream_config=upstream_config,
      log=log,
      environment_variables=self.cfg_serving_environment,
      version=_version,
      npy_shm_kwargs=npy_shm_kwargs,
      comm_method=comm_method,
    )
    
    self.log.stop_timer('create_server_new()')

    success = True
    if _server.inprocess:
      self.P("Model '{}' forced startup done".format(server_name), color='y')
      self.P("  Model '{}' serving process will NOT be running in parallel".format(server_name), color='y')
      self.P("  All logs will be deferred to current one.", color='y')
      self._servers[server_name] = {
        SMConst.SERVER: _server,
        # no need for comm but gonna put it anyway
        SMConst.COMM: comm_server,
        SMConst.PID : os.getpid(),
        SMConst.CREATED_DATE : self.log.now_str(nice_print=True, short=True),
      }
    else:
      self.owner.set_loop_stage('3.serving.start.{}.starting'.format(server_name))
      self.log.start_timer('create_server_start()')
      self.P("Running process {}...".format(_server))
      _server.daemon = True # TODO: must see if all servings work ok
      _server.start()
      # sleep 1-2 second to allow subprocess to "breath"
      sleep(2)
      _exit_code = _server.exitcode
      _pid = _server.pid
      self.owner.set_loop_stage('3.serving.start.{}.postsleep'.format(server_name))
      if _exit_code is not None:
        self.P("Serving process '{}' ended with code {}".format(server_name, _exit_code), color='r')
        comm_server.close()
        success = False
        if npy_shm is not None:
          npy_shm.shutdown()
      else:
        self.P("Serving process '{}' is running with PID {}".format(server_name, _pid))
        # now we wait for "READY" status
        self.P("  Waiting until '{}' responds".format(server_name))
        self.owner.set_loop_stage('3.serving.start.{}.wait'.format(server_name), is_dangerous=False)
        result = self._wait_for_result(
          server_name=server_name, 
          comm_eng=comm_server, 
          max_wait_time=float('inf'),
        )
        self.owner.set_loop_stage('3.serving.start.{}.received'.format(server_name))
        msg_type, msg_data = None, None
        if result is not None:          
          msg_type, msg_data = result
        if isinstance(msg_type, str) and msg_type.upper() == SMConst.READY:
          self.P("Received '{}' from '{}': {}".format(
            msg_type, server_name, msg_data), color='g')
          dct_server = {
            SMConst.SERVER: _server,
            SMConst.COMM: comm_server,
            SMConst.SHM: npy_shm,
            SMConst.PID: _pid,
            SMConst.CREATED_DATE : self.log.now_str(nice_print=True, short=False),
          }
          self._servers[server_name] = dct_server
          # now copy locally the remote configuration just to make sure all is good
          self._servers[server_name][SMConst.SERVER].config_model = self.get_server_config(server_name)
          self._servers[server_name][SMConst.SERVER].config_data = self._servers[server_name][SMConst.SERVER].config_model
        else:
          self._show_serving_error(server_name, result)
          self.P("Serving process '{}' FAILED. Closing pipes and process...".format(server_name), color='error')
          comm_server.close()
          _server.terminate()
          if npy_shm is not None:
            npy_shm.shutdown()
          self.P("  Serving process '{}' terminated: {}".format(server_name, _server), color='r')
          success = False
        # end if-else READY
      # end if-else exit_code is None
      self.log.stop_timer('create_server_start()')
    # end running in parallel

    if success:
      self.log.start_timer('create_server_parents()')
      # now lets populate similar "lower-level" servers if not already loaded
      similar_processes = _server.get_server_parents()
      covered_servers = _default_config.get(ct.COVERED_SERVERS, [])
      if server_class_name != server_name:
        covered_servers.append(server_class_name)
      self._maybe_complete_servers(server_name, similar_processes, covered_servers)
  
      # finally lets initialize server usage
      self._reset_usage(server_name)
      self.P("Server '{}' succesfully created.".format(server_name), color='g')
      self.show_servers()
      self.log.stop_timer('create_server_parents()')
      
    self.log.stop_timer('create_server')
    return server_name if success else None
  
  def _show_serving_error(self, server_name, result):
    # TODO: on every error deallocate parallel serving process
    server_name = self.get_server_name(server_name)
    if result is not None:
      if isinstance(result, dict):
        msg_data = result[SMConst.DATA]
        msg_type = result[SMConst.TYPE]            
      elif isinstance(result, tuple):
        msg_type, msg_data = result
      else:
        msg_type = "UNKNOWN ERROR"
        msg_data = str(result)
    else:
      msg_type = "Unknown status/error (NULL received)"
      msg_data = "Unknown status/error (NULL received) from " + server_name
    msg = "ServingManager: Received '{}' from model serving process '{}'".format(
      msg_type, 
      server_name
    )
    info = "  '{}' details:\n{}".format(msg_type, msg_data)
    self._create_notification(
      notif=ct.NOTIFICATION_TYPE.STATUS_EXCEPTION, 
      msg=msg, 
      info=info,
    )
    return


  def _create_servers(self):
    started = []
    for server_name in self._server_names:
      server_name = self.get_server_name(server_name)
      online = self._create_server(server_name) ### TODO upstream
      started.append((server_name, online))
    # done for each server
    return started

  def _send_message(self, server_name, command, inputs):
    server_name = self.get_server_name(server_name)
    self.__cmd_count += 1
    if self._servers[server_name].get(SMConst.SHM) is not None:
      self.log.start_timer('replace_ndarray_with_shm')
      processed_inputs = replace_ndarray_with_shm(inputs, self._servers[server_name][SMConst.SHM], debug=self.full_debug)
      self.log.stop_timer('replace_ndarray_with_shm')
    else:
      processed_inputs = inputs
    dct_cmd = {
      SMConst.TYPE : command,
      SMConst.DATA: processed_inputs,
      SMConst.SERVER_NAME : server_name,
      SMConst.CMD_ID : self.__cmd_count,
      SMConst.TIMESTAMP: time()
      }
    comm_eng = self._servers[server_name][SMConst.COMM]
    try:      
      if self.full_debug or command != SMConst.PREDICT:
        self.P('Sending {}:{} to {} on {}'.format(command, self.__cmd_count, server_name, comm_eng), color='y')
      self.log.start_timer('pipe_send_message')
      comm_eng.send(dct_cmd)
      self.log.stop_timer('pipe_send_message')
      if self.full_debug or command != SMConst.PREDICT:
        self.P('  Command {}:{} sent to {}'.format(command, self.__cmd_count, server_name), color='y')
        
    except:
      self.log.stop_timer('pipe_send_message') # double stop due to re-raise
      self.P("Error while sending command {} to server {}:{}".format(
         command, server_name, traceback.format_exc()), color='r')
      self.show_servers()
      raise ValueError('Comm pipe exception!')
    return
  
  
  def _wait_for_comm_message(self, server_name, comm_eng=None, max_wait_time=None):
    assert max_wait_time is not None, "ERROR: max_wait_time != None is required"
    server_name = self.get_server_name(server_name)
    if comm_eng is None:
      comm_eng = self._servers[server_name][SMConst.COMM]
    if max_wait_time is None:
      # we wait until we receive the answer
      payload = comm_eng.recv()
    else:
      # now we wait only `max_wait_time` seconds for the result
      has_payload = comm_eng.poll(max_wait_time)
      payload = None
      if has_payload:
        payload = comm_eng.recv()
    return payload
  
      
  def _cleanup_server(self, server_name, wait=False, terminate=False):    
    # cleanup and maybe even kill
    server_name = self.get_server_name(server_name)
    if server_name not in self._servers:
      return
    
    covered_by = self._servers[server_name].get(SMConst.COVERED_BY)
    if covered_by is not None:
      self.P("Attempted to cleanup the server '{}' covered by '{}'".format(
        server_name, covered_by), color='r'
      )
      return
    
    _server = self._get_server(server_name)
    self.P("Proceeding to serving process '{}' cleanup...".format(server_name), color='m')
    if _server.inprocess:
      self.P("  Parallel serving disabled. Cannot totally stop serving module '{}'".format(server_name), color='r')
      _server.shutdown()
      del _server
    else:
      if terminate and _server.is_alive():
        self.P("  Forcing '{}' termination: {}...".format(server_name, _server), color='m')
        _server.terminate()
      if wait:
        self.P("  Waiting for serving process '{}' ...".format(server_name), color='m')
        _server.join(5)
        if _server.exitcode is None:
          self.P("  Serving process '{}' did not respond to join()".format(server_name), color='r')
          _server.kill()
        #endif exitcode is None
      #endif wait  
      self.P("  Serving process status: {}".format(_server), color='m'
      )
      self.P("  Closing comm pipe for for serving process '{}' ...".format(server_name), color='m')
      self._servers[server_name][SMConst.COMM].close()
      if self._servers[server_name].get(SMConst.SHM) is not None:
        self._servers[server_name][SMConst.SHM].shutdown()
    
    del self._servers[server_name]
    remaining_servers = list(self._servers.keys())
    for svr in remaining_servers:
      if self._servers[svr].get(SMConst.COVERED_BY) == server_name:
        self.P("  Deleting covered server '{}' due to main server shutdown".format(svr), color='m')
        del self._servers[svr]
    return


  def _get_server(self, server_name):
    server_name = self.get_server_name(server_name)
    return self._servers[server_name][SMConst.SERVER] if server_name in self._servers else None

  def _close_server_on_null(self, server_name, max_wait_time=None):
    if max_wait_time is None:
      max_wait_time = self._server_wait_time(server_name)
    msg = "Server '{}' error: no message returned from server (maybe timeout) in over {}s.".format(
      server_name, max_wait_time)
    self.P(msg, color='r')
    self._create_notification(
      notif=ct.NOTIFICATION_TYPE.STATUS_EXCEPTION, 
      msg="'{}' server timeout".format(server_name), 
      info=msg,
      displayed=True,
    )
    self._cleanup_server(server_name, wait=True, terminate=True)
    return
  
  
  def _wait_for_result(self, server_name, message_callback=None, comm_eng=None, max_wait_time=None):
    """
    This method should ALWAYS be called with a default timeout. If the server does not send any information
    in the given time then something is wrong (including bad coding on the server side)
    """
    
    server_name = self.get_server_name(server_name)
    
    if self.full_debug:
      self.P(f'Waiting for data from {server_name}[timeouts: {self.get_timeouts(server_name)}| total: {self.total_timeouts}]', color='y')
      start_wait = time()
    #endif fulldebug

    if max_wait_time is None:
      max_wait_time = self._server_wait_time(server_name)

    result = self._wait_for_comm_message(
      server_name=server_name, comm_eng=comm_eng, max_wait_time=max_wait_time
    )

    if result is None:
      # something is clearly wrong so we cleanup and raise alarms    
      self._close_server_on_null(server_name=server_name, max_wait_time=max_wait_time)
      self.increment_timeouts(server_name=server_name)
      return None
    elif self.full_debug:
      msg_ts = result.get(SMConst.TIMESTAMP)
      elapsed_str = f'[Elapsed: {round(time() - msg_ts, 3)}]' if msg_ts is not None else ''
      self.P(f'  Received data {result.get(SMConst.CMD_ID)} from {server_name}{elapsed_str}', color='y')
      self.P(f'  Wait time: {round(time() - start_wait, 3)}', color='y')
    # end serving process problem
    
    # result is not none so we can get its data    
    msg_type = result[SMConst.TYPE].upper()  
    msg_data = result[SMConst.DATA]
    
    if SMConst.TIMERS in result and isinstance(result[SMConst.TIMERS], tuple):
      # we process timers if sent
      dct_timers, dct_timers_graph = result[SMConst.TIMERS]
      if dct_timers is not None and dct_timers_graph is not None:
        self.log.import_timers_section(
          dct_timers=dct_timers,
          dct_timers_graph=dct_timers_graph,
          section=str(server_name) + '_remote_main',
          overwrite=True,
        )
        self.P("Received timers from '{}' and updated remote section info".format(server_name), color='m')
    #endif received extra info --- timers
    
    # maybe during predict we might want to send intermediate messages before 'RESULT'
    # such as the ones that could be sent during training process
    # for this feature we will uses the `message_callback` for all non-result messages

    while msg_type != SMConst.RESULT:
      if msg_type == SMConst.ABNORMAL_EXIT:        
        self._show_serving_error(server_name, result)
        # if the serving process crashed then we wait for the process to finish
        # we can later try to re-create one
        self._cleanup_server(server_name, wait=True)
        return None
      else:
        self.P("Received '{}': from '{}' with data: '{}'".format(msg_type, server_name, msg_data), color='y')
        # now we can use message_callback in order to to custom processing on the received message if needed
        if message_callback is not None:
          message_callback(msg_data)
        else:
          # if no callback available and the message was not RESULT then we move on...
          break        
        # endif call the callback
      #endif abnormal or other message
      
      # periodic messages are multiplied by 2
      max_wait_time = self._server_wait_time(server_name) * 2
      result = self._wait_for_comm_message(server_name, comm_eng=comm_eng, max_wait_time=max_wait_time)
      if result is None:
        self._close_server_on_null(server_name=server_name, max_wait_time=max_wait_time)
        return None
      else:
        msg_type = result[SMConst.TYPE].upper()
        msg_data = result[SMConst.DATA]
      #endif valid result or not
    #endwhile not received the result
    return msg_type, msg_data

  def _increment_usage(self, server_name):
    server_name = self.get_server_name(server_name)
    if server_name not in self._servers:
      return None    
    dct_server = self._servers[server_name]
    if SMConst.PCOUNT not in dct_server:
      dct_server[SMConst.PCOUNT] = 0
    dct_server[SMConst.PCOUNT] += 1
    dct_server[SMConst.PTIME] = time()
    return dct_server[SMConst.PCOUNT]
  
  def _reset_usage(self, server_name):
    server_name = self.get_server_name(server_name)
    if server_name not in self._servers:
      return None    
    dct_server = self._servers[server_name]
    dct_server[SMConst.PCOUNT] = 0
    dct_server[SMConst.PTIME] = time()
    if SMConst.CREATED_DATE not in dct_server:
      dct_server[SMConst.CREATED_DATE] = self.log.now_str(nice_print=True, short=True)
    return
    
  
  
  #########################
  ###                   ###
  ###       PUBLIC      ###
  ###                   ###
  #########################
  
  def maybe_start_server(self, server_name, upstream_config=None, inprocess=False):
    result, server_class_name = None, None
    if isinstance(server_name, (list, tuple)):
      if len(server_name) != 2:
        self.P("Providing list/tuple instead of server name must be in (CLASS, NAME) format", color='error')
        return None
      server_class_name, _ = server_name # this extracts the actual serving process name
    server_name = self.get_server_name(server_name) # this constructs the instance name that can be <SERVING>_<INSTANCE>
    if server_name not in self._servers:
      self.P("`maybe_start_server`: '{}' server NOT available. Attempting to start server...".format(
        server_name), color='y')
      result = self._create_server(
        server_name=server_name,  # full name
        server_class_name=server_class_name, # just the serving process
        upstream_config=upstream_config,
        inprocess=inprocess
        )
      if result is None or result not in self._servers:
        self.__server_failures[server_name] += 1
        if self.__server_failures[server_name] > self.cfg_serving_environment.get(SMConst.SERVING_MAX_TRIES, SMConst.SERVING_MAX_TRIES_DEFAULT):
          # TODO: should this be a fatal error or should the node keep working?
          msg = "FATAL ERROR: Serving process creation failed for {} multiple times!".format(server_name)
          self.P(msg, color='r')
          self._create_notification(
            notif=ct.STATUS_TYPE.STATUS_EXCEPTION,          
            notif_code=ct.NOTIFICATION_CODES.SERVING_START_FATAL_FAIL,
            failed_ai_server=server_name,
            msg=msg,
            displayed=True,
            ct=ct,
          )
          self.P("Serving failure will generate a system shutdown!", color='r', boxed=True)
          raise ValueError(msg)        
        msg = "Serving process creation failed for {}!".format(server_name)
        self.P(msg, color='r')
        self._create_notification(
          notif=ct.STATUS_TYPE.STATUS_EXCEPTION,     
          notif_code=ct.NOTIFICATION_CODES.SERVING_START_FAILED,     
          failed_ai_server=server_name,
          msg=msg,
          displayed=True,
        )
        return None
      #endif

      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_NORMAL,
        notif_code=ct.NOTIFICATION_CODES.SERVING_START_OK,
        msg='Serving process {} creation succeeded'.format(result),
        ct=ct,
      )
    else:
      result = server_name
    #endif
    return result

  def is_covered_by(self, server_name):
    server_name = self.get_server_name(server_name)
    return self._servers[server_name].get(SMConst.COVERED_BY)

  def close_if_unused(self, server_name):
    server_name = self.get_server_name(server_name)
    if server_name not in self._servers:
      return False
    return self._servers[server_name][SMConst.SERVER].config_model.get('CLOSE_IF_UNUSED', False)

  def maybe_stop_unused_server(self, server_name):
    """
    This method will check if the server should be closed due to being unused.
    For this to have any effect the server must have
    Args:
      server_name: str or tuple, the server name to check

    Returns:

    """
    server_name = self.get_server_name(server_name)
    if server_name not in self._servers:
      return
    if self.is_covered_by(server_name) is not None:
      return
    if self.close_if_unused(server_name):
      self.P("Server '{}' is not used and should be stopped. Stopping...".format(server_name), color='y')
      self.stop_server(server_name)
    return

  def is_parallel(self, server_name):
    server_name = self.get_server_name(server_name)
    return not self._servers[server_name][SMConst.SERVER].inprocess
  
  def is_avail(self, server_name):
    server_name = self.get_server_name(server_name)
    return server_name in self._servers

  def get_server_config(self, server_name, key=None, local=False):
    """
    this function returns the serving process configuration. Will return current process
    retained config is local==True or the actual config if local==False
    
    
    """
    msg_type, msg_data, _config = None, None, None
    
    server_name = self.get_server_name(server_name)
    if server_name not in self._servers:
      self.P("Serving process '{}' is not active".format(server_name), color='r')
      return None

    _svr = self._servers[server_name][SMConst.SERVER]
    if _svr.inprocess:
      self.P("'{}' is running in current process. Local config is valid.".format(server_name))
      local = True
      
    if local:      
      _config = _svr.get_config()
    else:
      self._send_message(
        server_name=server_name, 
        command=SMConst.GET_CONFIG, 
        inputs=None
        )
      result = self._wait_for_result(server_name=server_name)      
      if result is not None:
        msg_type, msg_data = result
      # check if valid config message received
      if msg_type == SMConst.GET_CONFIG:
        _config = msg_data
      else:
        self._show_serving_error(server_name, result)
    #endif local or not
    if key is not None and _config is not None:
      return _config.get(key)
    return _config
  
  
  def start_server(self, server_name, upstream_config=None, inprocess=False):
    return self.maybe_start_server(
      server_name=server_name, 
      upstream_config=upstream_config,
      inprocess=inprocess
      )
  
  def stop_server(self, server_name):
    server_name = self.get_server_name(server_name)
    _server = self._get_server(server_name)
    if _server is None:
      # already stopped
      return
    covered_by = self.is_covered_by(server_name)
    if covered_by is not None:
      self.P("Serving process {} is covered by {}. Aborting shutdown".format(
        server_name,
        covered_by,
      ), color='m')
      return
    if _server is None:
      self.P("Serving process {} does not exist. Probably already stopped".format(server_name), color='r')
      return
    if not _server.inprocess:
      self.P("  Stopping serving process '{}' (PID:{}) ...".format(server_name, _server.pid), color='y')
      self._send_message(
        server_name=server_name, 
        command=SMConst.STOP, 
        inputs=None
      )
    # run cleanup no matter what
    self._cleanup_server(server_name, wait=True)
    if not _server.inprocess:
      self.P("  Serving process {} has been gracefully stopped.".format(server_name), color='y')
    return
  
  def stop_all_servers(self):
    self.P("Stopping all serving processes...", color='y')
    self.get_active_servers(show=True)
    gpu_info1 = self.log.gpu_info()
    if gpu_info1 is not None and len(gpu_info1) > 0:
      self.P("GPU0 free mem before shutdown: {} GB".format(gpu_info1[0]['FREE_MEM']), color='g')
    server_names = list(self._servers.keys())
    for server_name in server_names:      
      self.stop_server(server_name)
    gpu_info2 = self.log.gpu_info()
    if gpu_info2 is not None and len(gpu_info2) > 0:
      self.P("GPU0 free mem after shutdown: {} GB".format(gpu_info2[0]['FREE_MEM']), color='g')
    self.P("All servers stopped.", color='y')
    return
  
  def cleanup_stopped_servers(self):
    server_names = list(self._servers.keys())
    for server_name in server_names:
      _server = self._get_server(server_name)
      if _server.exitcode is not None:
        self.P("Serving process '{}' ended with exit code {}.".format(server_name, _server.exitcode))
        self._cleanup_server(server_name)
    return
  
  def get_plugin_default_config(self, signature):
    _module_name, _class_name, _cls_def, _config_dict = self._get_module_name_and_class(
      locations=ct.PLUGIN_SEARCH.LOC_SERVING_PLUGINS,
      name=signature,
      verbose=0,
      safety_check=True, # TODO: should we do this?
      suffix=ct.PLUGIN_SEARCH.SUFFIX_SERVING_PLUGINS,
      safe_locations=ct.PLUGIN_SEARCH.SAFE_LOC_SERVING_PLUGINS,     
      safe_imports=ct.PLUGIN_SEARCH.SERVING_SAFE_IMPORTS,
    )

    return _config_dict
  
  
  def get_predict_time(self, server_name):
    server_name = self.get_server_name(server_name)
    _server = self._get_server(server_name)
    if _server is not None:
      if _server.inprocess:
        timer_key = 'local_pred_' + server_name
      else:
        timer_key = 'remote_pred_' + server_name
    else:
      # if server is not active we assume it was remote and it stopped
      timer_key = 'remote_pred_' + server_name
    dct_timer = self.log.timers.get(timer_key)
    if dct_timer is not None:
      return dct_timer['MEAN']
    return None


  def simple_predict(self, server_name, inputs):
    res = None
    server_name = self.get_server_name(server_name)
    self.inprocess_predict = server_name
    self.inprocess_predict_start = time()
    self.log.start_timer('local_pred_' + server_name)
    try:
      _server = self._get_server(server_name)
      res = _server._in_process_predict(inputs)
    except Exception as e:
      info = str(e)
      self.P("Exception in '{}': {}".format(server_name, info), color='error')
      self._create_notification(
        notif=ct.NOTIFICATION_TYPE.STATUS_EXCEPTION,
        msg="Serving Manager exception for '{}' server.".format(server_name),
        info=info,
        displayed=True,
      )
      self.stop_server(server_name)
    self.log.stop_timer('local_pred_' + server_name)
    self.inprocess_predict = None
    return res
  
  
  def predict(self, server_name, inputs, message_callback=None, inprocess=False):
    """
    Single server predict

    Parameters
    ----------
    server_name : str or tuple
      name of the serving process.
      
    inputs : ndarray, dict, list, ...
      input data for the serving process.
      
    message_callback : func, optional
      callback that will be called when the server sends predict response message. The default is None.
    
    inprocess : bool
      if the serving process is not started then it will be started with inprocess or not

    Returns
    -------
    msg_data : list, dict, etc
      the result of the prediction/inference.

    """
    
    server_name = self.maybe_start_server(server_name, inprocess=inprocess)
    
    if server_name is None:
      return None
    
    self._increment_usage(server_name)
    
    msg_data = None # actually redundant
    if not self.is_parallel(server_name):      
      msg_data = self.simple_predict(server_name=server_name, inputs=inputs)
    else:
      # send the PREDICT message
      self.log.start_timer('remote_pred_' + server_name)
      self._send_message(
        server_name=server_name, # if this name is just a alias then the receiving process will figure that out
        command=SMConst.PREDICT,
        inputs=inputs,
        )
      # wait for the RESULT answer
      res = self._wait_for_result(server_name, message_callback=message_callback)
      self.log.stop_timer('remote_pred_' + server_name)
      if res is None:
        return None
      else:
        msg_type, msg_data = res
      if self.full_debug:
        self.P("Received from '{}' predict message '{}': {}".format(
          server_name, msg_type, type(msg_data)))
    
    self.collect_and_stop_idle_servers()
    return msg_data
  
  
  def server_runs_on_empty_input(self, server_name):
    svr = self._get_server(server_name)
    if svr is not None:
      # using svr.cfg_... is triky as the svr is initiated on different process
      # so we use the copy of config date received from parent
      runs_on_empty_input = svr.config_data['RUNS_ON_EMPTY_INPUT']
    else:
      runs_on_empty_input = None
      self.P("Attempted to call a inexisting server '{}'".format(
        self.get_server_config(server_name)
      ), color='r')
    return runs_on_empty_input
  
  def maybe_run_monitor(self):
    self._monitor_counter += 1
    if self._monitor_counter >= self._monitor_each_k_inferences:
      if self._monitor_callback is not None:
        self._monitor_callback()
      self._monitor_counter = 0
    return
  
  def check_blocked_inprocess_servers(self):
    """
    This method is designed for inprocesses only.
    Will monitor the `.predict` and force-throw a exception in main thread forcing 
    problematic models and simulating terminating a parallel process.
    """
    if self.inprocess_predict is None:
      return
    if not self.must_check_blocked_inprocess_serving:
      return
    elapsed = time() - self.inprocess_predict_start
    if elapsed > SMConst.MAX_INPROCESS_TIME:
      msg = "Forced inprocess server shutdown initiated for '{}' - predict blocked for {:.1}s > {:.1f}s".format(
        self.inprocess_predict,
        elapsed, SMConst.MAX_INPROCESS_TIME,
      )
      self.P(msg, color='error')
      self._create_notification(
        notif=ct.NOTIFICATION_TYPE.STATUS_EXCEPTION, 
        msg=msg,
        displayed=True,
      )
      ## Do the shutdown
      from naeural_core.utils.thread_raise import ctype_async_raise
      ctype_async_raise(self._main_tid, ValueError)
    # end must shutdown
    return
  
  
  def predict_parallel(self, dct_servers_inputs, inprocess=False):
    """
    Multi server predict

    Parameters
    ----------
    dct_servers_inputs : TYPE
      `server` : `inputs` dict.
      
      {
        'th_yolo_movenet' : [...]
        'th_iqa' : [...]
        ('th_second_stage', 'bascheti') : [...]
        ('th_second_stage', 'schiuri') : [...]
      }

    inprocess : TYPE, optional
      if a particular server is not running then start it with give `inprocess`. The default is False.

    Returns
    -------
    dct_resp : TYPE
      similar with the input, ie. a dict of `server:results` where results is a list (batch) of outputs 
      one output for each input (stream).

    """
    if self.log_timeouts_last_ts is None or time() - self.log_timeouts_last_ts > self.log_timeouts_period:
      self.log_timeouts_last_ts = time()
      msg_str = f"Serving servers timeouts: {self.total_timeouts}"
      details_str = '\n'.join([f"  {s_name}: {t_cnt}" for s_name, t_cnt in self._timeouts.items()])
      msg_str = msg_str + '\n' + details_str if len(details_str) > 0 else msg_str
      self.P(msg_str, color='y')
    #  endif log timeouts
    dct_resp = {}
    if not isinstance(dct_servers_inputs, dict):
      self.P("WARNING: parallel predict must receive a dict with at least one server and the associated inputs", color='r')
      return dct_resp

    ### disabled warnings in this case because in the main loop we can receive many warnings if no data is collected from the streams
    if len(dct_servers_inputs) == 0:
      self.collect_and_stop_idle_servers()
      return dct_resp

    self.log.start_timer('parallel_pred')
    
    self.log.start_timer('parallel_pred_filter_inputs')
    # this filtering has been moved from orchestrator >> {k:v for k,v in dct_servers_inputs.items() if len(v) > 0}
    dct_servers_inputs = {
      k:v for k,v in dct_servers_inputs.items() 
      if len(v) > 0 or self.server_runs_on_empty_input(k)
    }
    self.log.stop_timer('parallel_pred_filter_inputs')
    
    self.log.start_timer('parallel_pred_restart')

    self.owner.set_loop_stage('6.1.serve.run.parallel_pred_restart')
    # check if servers are online
    avail_servers = []
    for server_name in dct_servers_inputs:
      started = self.maybe_start_server(
        server_name,
        inprocess=inprocess
      )
      if started is not None:
        # while `started` can be different from `server_name` as in
        # `server_name == ('TH_MODEL','ONE')` and `started == 'TH_MODE_ONE'`
        # we must add upstream idenfier for upstream dict matching
        avail_servers.append(server_name)
        self._increment_usage(server_name)
    
    self.log.stop_timer('parallel_pred_restart')
    
    # start predicts: at this point we have to be carefull as if a server is (somehow) 
    # called more than once (separate entries in dct_servers_inputs and avail_servers)
    # we may have a crash at the first call that will invalidate the second call
    # thus we must make sure that we do not receive garabage from orchestrator !!!
    self.log.start_timer('parallel_pred_requests')

    for server_name in avail_servers:
      self.owner.set_loop_stage('6.1.serve.run.parallel_pred_requests.{}'.format(server_name))
      inputs = dct_servers_inputs[server_name]
      if self.is_parallel(server_name):
        self._send_message(
          server_name=server_name,
          command=SMConst.PREDICT,
          inputs=inputs,
        )
      else:
        self.log.start_timer('inprocess_pred_' + self.get_server_name(server_name))
        dct_resp[server_name] = self.simple_predict(
          server_name=server_name,
          inputs=inputs,
        )
        self.log.stop_timer('inprocess_pred_' + self.get_server_name(server_name))
    self.log.stop_timer('parallel_pred_requests')
    
    self.owner.set_loop_stage('6.1.serve.run.maybe_run_monitor')
    self.maybe_run_monitor()

    # collect answers if servers are not inprocess
    self.log.start_timer('parallel_pred_collect')
    self.owner.set_loop_stage('6.1.serve.run.parallel_pred_collect')
    for server_name in avail_servers:
      if not self.is_avail(server_name):
        # not available anymore!
        self.P("Serving process '{}' closed during predict".format(server_name), color='r')
        dct_resp[server_name] = None
        continue
      if self.is_parallel(server_name):
        # if parallel then sync check (wait) for result
        self.owner.set_loop_stage('6.1.serve.run.parallel_pred_collect.' + self.get_server_name(server_name))
        max_wait_time = self._server_wait_time(server_name)
        res = self._wait_for_result(
          server_name=server_name, 
          max_wait_time=max_wait_time,
        )
        dct_resp[server_name] = res[1] if res is not None else None
      else:
        # do nothing if not parallel as the result is already collected
        pass 
    self.log.stop_timer('parallel_pred_collect')
    # done all predicts
    
    self.owner.set_loop_stage('6.1.serve.run.stop_idle_servers')
    self.collect_and_stop_idle_servers(avail_servers)

    self.log.stop_timer('parallel_pred')
    return dct_resp
  
  
  def get_predict_parallel_time(self):
    if 'parallel_pred' in self.log.timers:
      return self.log.timers['parallel_pred']['MEAN']
    return None
  
  
  
  def collect_and_stop_idle_servers(self, last_servers_run=None, force_collection=False):
    if not force_collection:
      if self._collector_call_counter < self._server_collector_delay:
        self._collector_call_counter += 1
        return
      else:
        self._collector_call_counter = 0
      
    self.log.start_timer('idle_shutdown')
    current_servers = list(self._servers.keys())
    for server_name in current_servers:
      if server_name not in self._servers:
        # already deleted within the loop :)
        continue
      if self._servers[server_name].get(SMConst.COVERED_BY) is not None:
        # this is a alias server covered by someone else
        continue
      count = self._servers[server_name][SMConst.PCOUNT]
      last_time = self._servers[server_name][SMConst.PTIME]
      delay = time() - last_time
      if delay > self._server_collector_timedelta:
        msg1 = "'{}' server forced shutdown initiated".format(server_name)          
        msg2 = "WARNING: Serving process '{}' idle {:.1f}s > {}s (usage: {}). {}Shutting down...".format(
          server_name, delay, self._server_collector_timedelta, count,
          "Last `parallel_predict` on {}. ".format(self.log.time_to_str(last_time)), # if last_servers_run is not None else '',
          )
        self.P(msg2, color='r')
        if self._server_collector_timedelta < 3600:
          msg1 = msg1 + ' - SERVER_COLLECTOR_TIMEDELTA:{} may be too low!'.format(self._server_collector_timedelta)
          self.P("  " + msg1, color='r')
        self._create_notification(
          notif=ct.NOTIFICATION_TYPE.STATUS_EXCEPTION, 
          msg=msg1, 
          info=msg2,
        )
        self.get_active_servers(show=True)
        self.stop_server(server_name)
        self.show_servers('Post server-cleanup serving process status:')
    self.log.stop_timer('idle_shutdown')
    return


  def keep_active_servers(self, active_server_names, inprocess=True):
    active_server_names = [self.get_server_name(x) for x in active_server_names]
    
    lst_in_use = self.get_active_server_names()
    lst_start = list(set(active_server_names) - set(lst_in_use))
    lst_stop = list(set(lst_in_use) - set(active_server_names))
    
    if lst_start:
      self.P('Found {} new graphs to start: {}'.format(len(lst_start), lst_start))
      for name in lst_start:
        self.start_server(
          server_name=name,
          inprocess=inprocess
          )
      
    if lst_stop:
      self.P('Found {} graphs to deallocate: {}'.format(len(lst_stop), lst_stop))
      for name in lst_stop:
        self.stop_server(name)
    return
  
  def get_active_server_names(self):
    return list(self._servers.keys())

  def get_active_servers(self, show=True, color='b'):
    res = []
    lines = []
    if show:
      self.P("Analysing current active serving processes...", color=color)
      if len(self._servers) == 0:
        self.P("  -- No active servers running --", color=color)
      else:
        lines = ["==== List of active serving processes ===="]
    
    server_names = list(self._servers.keys())    
    for server_name in server_names:
      count = self._servers[server_name].get(SMConst.PCOUNT, -1)
      covered_by = self._servers[server_name].get(SMConst.COVERED_BY)
      last_time = self._servers[server_name].get(SMConst.PTIME, 0)
      when_created = self._servers[server_name].get(SMConst.CREATED_DATE, "")
      str_last_run_gmt = strftime('%H:%M:%S', gmtime(last_time))
      str_last_run_local = self.log.time_to_str(last_time)
      str_last_run = '{} / GMT: {}'.format(str_last_run_local, str_last_run_gmt)
      idle_time = (time() - last_time) if last_time > 0 else -1
      svr_name = self._get_server(server_name).__class__.__name__
      if covered_by is None:
        res.append(OrderedDict({
          'Name (ID)' : server_name,
          'Server'    : svr_name,
          'Created'   : when_created,
          'Last run'  : str_last_run_local,
          'Idle (s)'  : round(idle_time, 1),
          'Run count' : count,
          'Inprocess' : self._get_server(server_name).inprocess,
          })
        )
      if show:
        lines.append("Server '{}':{}".format(
          server_name, " covered by {}:{}".format(
            covered_by, 
            self._servers[covered_by][SMConst.SERVER] if covered_by in self._servers else '<deallocated>'
            ) if covered_by is not None else '',
          ))
        lines.append("  Process:  {}".format(svr_name))
        lines.append("  Created:  {}".format(when_created))
        lines.append("  Last run: {}".format(str_last_run))
        lines.append("  Run cnt:  {}".format(count))
        lines.append("  Idle/Max: {:.0f}s/{:.0f}s".format(idle_time, self._server_collector_timedelta))
        self.P("\n".join(lines), color=color)
        lines = []
    return res
  
  def show_servers(self, title="Current servers:", color='b'):
    lst_log = [title]
    if len(self._servers) == 0:
      lst_log.append("  -- No servers running to show --")
    for svr in self._servers:
      lst_log.append("  {}:".format(svr))
      for k,v in self._servers[svr].items():
        lst_log.append("    {:<15}{}".format(k + ':', v))
    full_log = "\n".join(["    " + x for x in lst_log])
    self.P(full_log.lstrip(), color=color)
    return




from naeural_core.business.base import BasePluginExecutor as BasePlugin



_CONFIG = {
  **BasePlugin.CONFIG,
  
  "PROCESS_DELAY"           : 5,
  "ALLOW_EMPTY_INPUTS"      : True,
  
  "SEND_MANIFEST_EACH"      : 60,
  
  'RESULT_KEY'              : 'exec_result',
  'ERROR_KEY'               : 'exec_errors',  
  'WARNING_KEY'             : 'exec_warning',
  

  'VALIDATION_RULES' : {
    **BasePlugin.CONFIG['VALIDATION_RULES'],

  },  
}


class CodeExecutorPlugin(BasePlugin):
  CONFIG = _CONFIG

  def on_init(self):
    self.__last_ready_timestamp = 0
    self.P(f"{self.__class__.__init__} initialized and sending manifest each {self.cfg_send_manifest_each} s.")
    return
  
  def __exec_code(self, base64code, **kwargs):
    result, errors, warnings = None, None, None
    if base64code is not None and len(base64code) > 0:
      res = self.exec_code(
        str_b64code=base64code,
        debug=self.cfg_debug_exec,
        self_var='plugin', # we set plugin local for exec code
        modify=True,
      )
      if isinstance(res, tuple):
        result, errors, warnings = res
    else:
      errors = "Received empty CODE"

    kwargs = {
      self.cfg_result_key : result,
      self.cfg_error_key : errors,
      self.cfg_warning_key : warnings,
    }            
    payload = self._create_payload(
      **kwargs,
    )    
    
    if len(warnings) > 0:
      self.P("Execution warnings: {}".format(warnings), color='r')    
      
    if errors is not None:
      self.P("Execution errors: {}".format(errors), color='r')    
    return
  
  def on_command(self, data, **kwargs):
    str_b64code = data.get('CODE')
    self.__exec_code(b64code=str_b64code, **kwargs)
    return
  
  def process(self):
    payload = None
    if (self.time() - self.__last_ready_timestamp) > self.cfg_send_manifest_each:
      self.P("Sending ready status...")
      payload = self._create_payload(
        status="Ready to receive commands."
      )
      self.__last_ready_timestamp = self.time()
    #endif send manifest
    return payload
    
      



from naeural_core import DecentrAIObject
from naeural_core import Logger
import json
from naeural_core.serving.ai_engines.utils import (
  get_serving_process_given_ai_engine,
  get_ai_engine_given_serving_process,
  get_params_given_ai_engine
)

class MainLoopDataHandler(DecentrAIObject):
  """
  In order to understand the methods of this class please study below data inputs:

  ######################
  1.`self._dct_captures`
  ######################
  `self._dct_captures` is the data collected by CaptureThreadManager:
    {
      'simple_image_stream' : { ### th_y5l6s -> omni: lpr
        'STREAM_NAME' : 'simple_image_stream',
        'STREAM_METADATA' : {Dictionary with general stream metadata},
        'INPUTS' : [
          {
            'IMG' : np.ndarray,
            'STRUCT_DATA' : None,
            'INIT_DATA' : None,
            'TYPE' : 'IMG',
            'METADATA' : {Dictionary with current input metadata}
          },
        ]
      },

      'multi_modal_2_images_one_sensor_stream' : { ### th_y5l6s, anomaly_detection -> anomaly_detection, omni: pose, safety_gear, detection
        'STREAM_NAME' : 'multi_modal_2_images_one_sensor_stream',
        'STREAM_METADATA' : {Dictionary with general stream metadata},
        'INPUTS' : [
          {
            'IMG' : np.ndarray,
            'STRUCT_DATA' : None,
            'INIT_DATA' : None,
            'TYPE' : 'IMG',
            'METADATA' : {Dictionary with current input metadata}
          },

          {
            'IMG' : np.ndarray,
            'STRUCT_DATA' : None,
            'INIT_DATA' : None,
            'TYPE' : 'IMG',
            'METADATA' : {Dictionary with current input metadata}
          },

          {
            'IMG' : None,
            'STRUCT_DATA' : an_object,
            'INIT_DATA' : None,
            'TYPE' : 'STRUCT_DATA',
            'METADATA' : {Dictionary with current input metadata}
          },
        ]
       }
    }

    N.B. Each dictionary in 'INPUTS' is defined as below:
      {
        'IMG' : ..., # 'IMG' and 'STRUCT_DATA' are mutual exclusive (one of them should be None)
        'STRUCT_DATA' : ..., # 'IMG' and 'STRUCT_DATA' are mutual exclusive (one of them should be None)
        'INIT_DATA' : ..., # could be None if no initial data is provided
        'TYPE' : ..., # this value should be 'IMG' or 'STRUCT_DATA' in order to specify which field is completed
        'METADATA' : {Dictionary with current input metadata}
      }

  ###############################
  2.`self._dct_instances_details`
  ###############################
  `self._dct_instances_details` represents the instance details taken from PluginsManager
    (see PluginsManager -> property `dct_instances_details`)
    It is a mapping between each current plugin instance and their details. Each instance hash is created based on
    stream_name, signature and instance_id 
    (see  `instance_hash = self.log.hash_object((stream_name, signature, instance_id))` in PluginsManager)

  {
    'instance_hash_1' : (stream_id, signature, instance_config),
    'instance_hash_2' : (stream_id, signature, instance_config),
    ....
  }


  #############################################
  3. `self._dct_serving_processes_details`
  #############################################

  IMPORTANT:
    
    `self._dct_serving_processes_details` represents the serving processes details 
    taken from PluginsManager, i.e. the serving upstream config delivered from biz plugins
    via 'INFERENCE_AI_ENGINE_PARAMS' and used during inferences. This is different from the 
    serving startup upstream config delivered from 'STARTUP_AI_ENGINE_PARAMS'
  
    (see PluginsManager -> attribute `dct_serving_processes_details`)

  {
    'th_y5l6s' : {
      ('simple_image_stream', dict_upstream_yolo_model_params_1) : ['instance_hash_1', 'instance_hash_3'],
      ('multi_modal_2_images_one_sensor_stream', dict_upstream_yolo_model_params_1) : ['instance_hash_2'],
    },

    'anomaly_detection_model' : {
      ('multi_modal_2_images_one_sensor_stream', dict_upstream_anomaly_model_params_1) : ['instance_hash_2'],
    }
  }

  # {
  #   'y5_omni' : {
  #     ('simple_image_stream', dict_upstream_yolo_model_params_1) : ['instance_hash_1', 'instance_hash_3'],
  #     ('multi_modal_2_images_one_sensor_stream', dict_upstream_yolo_model_params_1) : ['instance_hash_2'],
  #   }
  # }


  """

  def __init__(self, log : Logger, owner, **kwargs):
    self._dct_models_stream_idx = None
    self.owner = owner

    self._dct_captures = None
    self._dct_instances_details = None
    self._dct_serving_processes_details = None

    self.dct_business_inputs = None
    super(MainLoopDataHandler, self).__init__(log=log, prefix_log='[DAGGM]', **kwargs)
    return

  def update(self,
             dct_captures,
             dct_instances_details,
             dct_serving_processes_details):
    """
    Updates the internal state of the manager.

    Parameters:
    ----------
    dct_captures:
      The data collected by CaptureThreadManager

    dct_instance_details:
      Instance details taken from PluginsManager

    dct_model_serving_processes_details:
      Model serving processes details taken from PluginsManager
    """

    self._dct_captures = dct_captures
    self._dct_instances_details = dct_instances_details
    self._dct_serving_processes_details = dct_serving_processes_details
    return

  def _get_stream_captured_data(self, stream_name):
    if stream_name not in self._dct_captures:
      return {}

    return {
      'STREAM_NAME'     : self._dct_captures[stream_name]['STREAM_NAME'],
      'STREAM_METADATA' : self._dct_captures[stream_name]['STREAM_METADATA'],
      'INPUTS'          : self._dct_captures[stream_name]['INPUTS'],
    }

  def append_captures(self):
    self.dct_business_inputs = {}

    for instance_hash, (stream_name, _, _) in self._dct_instances_details.items():
      self.dct_business_inputs[instance_hash] = self._get_stream_captured_data(stream_name)
    #endfor

    """
    At this point `self.dct_business_inputs` will be like this:
    
    {
      'instance_hash_1' : {
        'STREAM_NAME' : ...,
        'STREAM_METADATA' : ...,
        'INPUTS' : ...
      },
      
      'instance_hash_2' : {
        'STREAM_NAME' : ...,
        'STREAM_METADATA' : ...,
        'INPUTS' : ...
      },
      
      ...
    }
    """

    return

  def aggregate_for_inference(self):
    """
    This function is the main point where the input data is prepared for each individual
    serving process. It combines the following data:
      - input data from DCTs using `_get_stream_captured_data`
      - serving params pre-defined for inference time by AI_ENGINE using `get_params_given_ai_engine`
      - inference time params from biz plugin `INFERENCE_AI_ENGINE_PARAMS` param passed from
        `BizMgr.dct_serving_processes_details` via local `_dct_serving_processes_details` as JSON string
        that will be decoded into the actual params dict. 
        IMPORTANT: 
          while `STARTUP_AI_ENGINE_PARAMS` are passed directly into serving `config_data` of the serving 
          process, now the inference-time upstream config is given in input data and thus its extraction 
          must be handled by the serving process

    Returns
    -------
    dict with all the inputs.

    """

    self._dct_models_stream_idx = {}
    dct_servers_inputs = {}
    
    for ai_engine, dct in self._dct_serving_processes_details.items():
      # TODO:
      #   AI_ENGINE handling should be done in ServingManager
      serving_process = get_serving_process_given_ai_engine(ai_engine)
      # next we get params from AI ENGINE config
      ai_engine_params = get_params_given_ai_engine(ai_engine)
      dct_servers_inputs[serving_process] = []
      self._dct_models_stream_idx[serving_process] = {}
      i = 0
      for _, (stream, biz_plugin_model_params_json) in enumerate(list(dct.keys())):
        server_input = self._get_stream_captured_data(stream)
        ### make sure that data was collected on this particular stream
        if bool(server_input) or self.owner.serving_manager.server_runs_on_empty_input(serving_process): 
          # order is important - biz plugin should overwrite AI_ENGINE params
          server_input['SERVING_PARAMS'] = {**ai_engine_params, **json.loads(biz_plugin_model_params_json)}
          dct_servers_inputs[serving_process].append(server_input)
          self._dct_models_stream_idx[serving_process][i] = (stream, biz_plugin_model_params_json, ai_engine_params)
          i += 1
        #endif
      #endfor
    #endfor

    """
    At this point, `dct_servers_inputs` will be like this:
    
    {
      'th_y5l6s' : [
        {'STREAM_NAME' : 'simple_image_stream', 'STREAM_METADATA' : ..., 'INPUTS' : ..., 'SERVING_PARAMS' : dict_upstream_yolo_model_params_1},
        {'STREAM_NAME' : 'multi_modal_2_images_one_sensor_stream', 'STREAM_METADATA' : ..., 'INPUTS' : ..., 'SERVING_PARAMS' : dict_upstream_yolo_model_params_1}
      ],
      
      'anomaly_detection_model' : [
        {'STREAM_NAME' : 'multi_modal_2_images_one_sensor_stream', 'STREAM_METADATA' : ..., 'INPUTS' : ..., 'SERVING_PARAMS' : dict_upstream_anomaly_model_params_1}
      ]
    }    
    """

    return dct_servers_inputs ## this object will be input for ServingManager's `predict_parallel`

  def append_inferences(self, dct_models_outputs):
    """
    
    TODO: this does not correctly support multiple instances of the same serving process
    
    Parameters:
    ----------
    dct_model_outputs:
      The response from ServingManager's `predict_parallel` (see ModelServingProcess's `pack_results`)
    """

    """
    `dct_models_outputs` :
    {
      'th_y5l6s' : {
        'INFERENCES_META' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : ...},
        'INFERENCES' : [ ### 'INFERENCES' will have the same length as the input for 'th_y5l6s'  - below example model input = 3 imgs
          [
            [{'TLBR_POS' : ...}, {'TLBR_POS' : ...}]
          ],
          [ ### this list has two elements because on 'multi_modal_2_images_one_sensor_stream', 'th_y5l6s' dealed with 2 images!!!
            [{'TLBR_POS' : ...}, {'TLBR_POS' : ...}, {'TLBR_POS' : ...}],
            [{'TLBR_POS' : ...}]
          ]
        ]
      },
      
      'anomaly_detection_model' : {
        'INFERENCES_META' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : ...},
        'INFERENCES' : [ ### 'INFERENCES' will have the same length as the input for 'anomaly_detection_model' 
          [
            'True/False' ## the output of the 'anomaly_detection_model'
          ]
        ]
      }
      
    }    
    """

    for serving_process, dct_output in dct_models_outputs.items():
      if dct_output is None:
        continue
      for i, model_inference in enumerate(dct_output['INFERENCES']):
        stream, biz_plugin_model_params_json, ai_engine_params = self._dct_models_stream_idx[serving_process][i]
        ai_engine = get_ai_engine_given_serving_process(
          serving_process=serving_process, 
          params=ai_engine_params
        )
        key = (stream, biz_plugin_model_params_json)
        suited_instances_for_crt_inference = self._dct_serving_processes_details[ai_engine][key]
        for instance in suited_instances_for_crt_inference:
          if 'SERVING_PARAMS' not in self.dct_business_inputs[instance]:
            self.dct_business_inputs[instance]['SERVING_PARAMS'] = {}
          if 'INFERENCES' not in self.dct_business_inputs[instance]:
            self.dct_business_inputs[instance]['INFERENCES'] = {}
          if 'INFERENCES_META' not in self.dct_business_inputs[instance]:
            self.dct_business_inputs[instance]['INFERENCES_META'] = {}

          self.dct_business_inputs[instance]['SERVING_PARAMS'][ai_engine] = json.loads(biz_plugin_model_params_json)
          self.dct_business_inputs[instance]['INFERENCES'][ai_engine] = model_inference
          self.dct_business_inputs[instance]['INFERENCES_META'][ai_engine] = dct_output['INFERENCES_META']
        # endfor - each business plugin that should receive the input from the current inference
      # endfor - each inference for current model
    # endfor - each model with its inferences

    """
    At this point `self.dct_business_inputs` will be like this
    (will have completed the 'INFERENCES' only for the instances that do not have configured 'MODEL_SERVING_PROCESS' as None):

    {
      'instance_hash_1' : {
        'STREAM_NAME' : ...,
        'STREAM_METADATA' : ...,
        'INPUTS' : ...,
        'INFERENCES' : {
          'th_y5l6s' : [[{'TLBR_POS' : ...}, {'TLBR_POS' : ...}]],
        },
        'INFERENCES_META' : {
          'th_y5l6s' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : ...}
        }
      },

      'instance_hash_2' : {
        'STREAM_NAME' : ...,
        'STREAM_METADATA' : ...,
        'INPUTS' : ...,
        'INFERENCES : {
          'th_y5l6s' : [
            [{'TLBR_POS' : ...}, {'TLBR_POS' : ...}, {'TLBR_POS' : ...}],
            [{'TLBR_POS' : ...}]
          ],
          
          'anomaly_detection_model' : [
            'True/False'
          ]
        },
        'INFERENCES_META' : {
          'th_y5l6s' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : ...},
          'anomaly_detection_model' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : ...}
        }
      },
      
      'instance_hash_3' : {
        'STREAM_NAME' : ...,
        'STREAM_METADATA' : ...,
        'INPUTS' : ...,
        'INFERENCES : {
          'th_y5l6s' : [[{'TLBR_POS' : ...}, {'TLBR_POS' : ...}]],
        },
        'INFERENCES_META' : {
          'th_y5l6s' : {'SYSTEM_TYME' : ..., 'VER' : ..., 'PICKED_INPUT' : ...}
        }
      }
    }
    """
    return



import torch as th
import numpy as np
import abc
import torchvision.transforms as T
from typing import List, Tuple, Any
from naeural_core import DecentrAIObject
from naeural_core import Logger
from naeural_core.local_libraries.nn.th.trainer import ModelTrainer
from naeural_core.local_libraries.nn.th.training_utils import to_device

class TrainingCallbacks(DecentrAIObject):

  def __init__(self, log : Logger, owner : ModelTrainer,
               preprocess_before_fw_callback,
               training_device : str = 'cuda:0',
               level_analysis=1,
               dct_augmentation_plan=None,
               **kwargs):
    self._owner = owner
    self._level_analysis = level_analysis
    self._preprocess_before_fw_callback = preprocess_before_fw_callback
    self._dct_augmentation_plan = dct_augmentation_plan or {}
    self._augmentation = None

    self.training_device = training_device
    super(TrainingCallbacks, self).__init__(log=log, **kwargs)
    return

  def startup(self):
    super().startup()
    self.__name__ = self.log.name_abbreviation(self.__class__.__name__)
    self._get_augmentation()
    return

  def P(self, s, **kwargs):
    super().P(s, prefix=True, **kwargs)
    return

  @to_device('training_device')
  def _augment_input_callback(self, x, epoch=None):
    if epoch in self._dct_augmentation_plan:
      self._get_augmentation(epoch=epoch)

    return self._augmentation(x)

  def _get_augmentation(self, epoch=None):
    kwargs = {}
    if epoch is not None:
      kwargs = self._dct_augmentation_plan[epoch]

    lst_augmentations_definitions = self._lst_augmentations(**kwargs)
    lst_augmentations = [x[0](**x[1]) for x in lst_augmentations_definitions]
    self._augmentation = T.Compose(lst_augmentations)
    return

  @property
  def _model(self):
    return self._owner.model

  def _get_idx(self, lst_indexes : List[np.ndarray]) -> np.ndarray:
    idx = np.hstack(lst_indexes)
    return idx

  @abc.abstractmethod
  def _get_y(self, lst_y):
    """
    Method that concatenates batches of targets into a List[Tensor] or Tensor (depending whether the model is
    multi-output or single-output) that is provided to evaluation / testing callback.

    The forward pass is performed in batches. For each batch, the target is put in a list.
    After the full forward pass, that list (`lst_y`) should be concatenated to be afterwards used in the
    evaluation / testing callback

    Parameters
    ----------
    lst_y: List[Tensor] for single-output or List[List[Tensor]] for multi-output
      List of target for each batch.

    Returns
    -------
    y: Tensor for single-output or List[Tensor] for multi-output
      The concatenated target
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _get_y_hat(self, lst_y_hat):
    """
    Method that concatenates batches of inference into a List[Tensor] or Tensor (depending whether the model is
    multi-output or single-output) that is provided to evaluation / testing callback.

    The forward pass is performed in batches. For each batch, the inference is put in a list.
    After the full forward pass, that list (`lst_y_hat`) should be concatenated to be afterwards used in the
    evaluation / testing callback

    Parameters
    ----------
    lst_y_hat: List[Tensor] for single-output or List[List[Tensor]] for multi-output
      List of inference for each batch.

    Returns
    -------
    y_hat: Tensor for single-output or List[Tensor] for multi-output
      The concatenated inference
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _evaluate_callback(self,
                         epoch : int,
                         dataset_info : dict,
                         y, y_hat,
                         idx : np.ndarray,
                         key: str = 'dev') -> dict:
    """
    The evaluation method that is called during training after each epoch.

    Parameters
    ----------
    epoch: int
      At which epoch the evaluation is performed

    dataset_info: dict
      Reference to DataLoaderFactory's `self.dataset_info`

    y : Tensor for single-output or List[Tensor] for multi-output
      The target that should be evaluated with the model's output.

    y_hat: Tensor for single-output or List[Tensor] for multi-output
      The model output that should be evaluated with the target

    idx: np.ndarray
      Indexes in `dataset_info` for `y` and `y_hat`

    key: str
      Which dataset is evaluated

    Returns
    -------
    dict{str:float}
      A dictionary {metric_name: value}
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _test_callback(self,
                     dataset_info : dict,
                     y, y_hat,
                     idx : np.ndarray) -> dict:
    """
    The test method that is called at the end of the training

    Parameters
    ----------

    dataset_info: dict
      Reference to DataLoaderFactory's `self.dataset_info`

    y : Tensor for single-output or List[Tensor] for multi-output
      The target that should be evaluated with the model's output.

    y_hat: Tensor for single-output or List[Tensor] for multi-output
      The model output that should be evaluated with the target

    idx: np.ndarray
      Indexes in `dataset_info` for `y` and `y_hat`

    Returns
    -------
    dict{str:float}
      A dictionary {metric_name: value}
    """
    raise NotImplementedError

  @abc.abstractmethod
  def _lst_augmentations(self, **kwargs) -> List[Tuple[Any, dict]]:
    """
    List of augmentation techniques that will be performed for the training dataset.

    This augmentation step is performed between "load preprocess" and "forward preprocess" (please see
    libraries.nn.th.trainin.data.base, methods `_lst_on_load_preprocess` and `_lst_right_before_forward_preprocess`)

    All the augmentation techniques should be wrapped in classes that expose a __call__ method.
    (please see libraries.nn.th.image_dataset_stage_preprocessor)

    All these augmentation methods are automatically packed into a th transformation (T.Compose)

    Returns:
    -------
    List[Tuple[Any, dict]]
      for each augmentation technique, a tuple with: a) the class reference and b) the class __init__ kwargs
      should be provided
    """
    raise NotImplementedError

  def _forward_and_collect(self, data_generator: th.utils.data.DataLoader, dataset_info : dict):
    with th.no_grad():
      steps_per_epoch = len(data_generator)
      iterator = iter(data_generator)
      lst_y, lst_y_hat, lst_indexes = [], [], []
      for i in range(steps_per_epoch):
        x_batch, y_batch, idx_batch = next(iterator)
        x_batch, y_batch = self._transform_at_eval_time(x_batch, y_batch)
        model_result = self._model(x_batch)
        if isinstance(model_result, th.Tensor):
          lst_y_hat.append(model_result.cpu())
          lst_y.append(y_batch.cpu())
        else:
          lst_y_hat.append([_y.cpu() for _y in model_result])
          lst_y.append([_y.cpu() for _y in y_batch])
        #endif
        lst_indexes.append(idx_batch)

        if self.log.is_main_thread:
          print(
            '\r  Forward progress: {}/{} ({:.2f}%)'.format(i + 1, steps_per_epoch, 100 * (i + 1) / steps_per_epoch),
            flush=True, end=''
          )
      #endfor
      y = self._get_y(lst_y)
      y_hat = self._get_y_hat(lst_y_hat)
      idx = self._get_idx(lst_indexes)
    #endwith
    return y, y_hat, idx


  def evaluate_callback(self, epoch : int, data_generator : th.utils.data.DataLoader, dataset_info : dict, key : str = 'dev', **kwargs):
    self._model.eval()
    self.P("START EVALUATION for: {}".format(key), color='g')
    y, y_hat, idx = self._forward_and_collect(data_generator=data_generator, dataset_info=dataset_info)
    dct_score = self._evaluate_callback(
      epoch=epoch,
      dataset_info=dataset_info,
      y=y, y_hat=y_hat, idx=idx, key=key
    )
    self.P("END EVALUATION for: {}".format(key), color='g')
    self._model.train()
    return dct_score

  def test_callback(self, data_generator : th.utils.data.DataLoader, dataset_info : dict, **kwargs):
    self._model.eval()
    self.P("START TESTING", color='g')
    y, y_hat, idx = self._forward_and_collect(data_generator=data_generator, dataset_info=dataset_info)
    dct_score = self._test_callback(
      dataset_info=dataset_info,
      y=y, y_hat=y_hat, idx=idx
    )
    self.P("END TESTING", color='g')
    self._model.train()
    return dct_score

  def dev_callback(self, data_generator : th.utils.data.DataLoader, dataset_info : dict, **kwargs):
    self._model.eval()
    self.P("START DEV TESTING", color='g')
    y, y_hat, idx = self._forward_and_collect(data_generator=data_generator, dataset_info=dataset_info)
    dct_score = self._dev_callback(
      dataset_info=dataset_info,
      y=y, y_hat=y_hat, idx=idx
    )
    self.P("END DEV TESTING", color='g')
    self._model.train()
    return dct_score


  def _transform_at_training_time(self, x, y, epoch=None):
    x = self._augment_input_callback(x, epoch=epoch)
    x, y = self._preprocess_before_fw_callback(x, y)
    return x, y

  def _transform_at_eval_time(self, x, y):
    x, y = self._preprocess_before_fw_callback(x, y)
    return x, y

  def _get_batch_lst_x_lst_y(self, batch, y_index, batch_index):
    if batch_index:
      batch = batch[:-1]

    if len(batch) == 1:
      lst_th_x = [batch[0]]
      lst_th_y = None
    else:
      lst_th_x = batch[:y_index]
      lst_th_y = batch[y_index:]
    #endif

    if len(lst_th_x) == 1:
      lst_th_x = lst_th_x[0]

    if len(lst_th_y) == 1:
      lst_th_y = lst_th_y[0]

    return self._transform_at_training_time(lst_th_x, lst_th_y, epoch=self._owner.current_epoch)

  def train_forward(self, model, lst_th_x):
    return model(lst_th_x)

  def train_on_batch_callback(self, model, optimizer, losses, batch, y_index=1, batch_index=False):
    lst_th_x, lst_th_y = self._get_batch_lst_x_lst_y(batch=batch, y_index=y_index, batch_index=batch_index)

    # maybe move data on device if default is disabled
    if isinstance(lst_th_x, (list, tuple)):
      data_device = lst_th_x[0].device
    else:
      data_device = lst_th_x.device
    #endif

    model_device = next(model.parameters()).device
    assert data_device == model_device

    th_yh = self.train_forward(model, lst_th_x)
    losses_vals = []

    if not isinstance(th_yh, (list, tuple)):
      th_yh = [th_yh]

    if not isinstance(lst_th_y, (list, tuple)):
      lst_th_y = [lst_th_y]

    if len(losses) != len(lst_th_y):
      assert len(losses) == 1

    if lst_th_y is not None:
      ### TODO: CHECK If ok
      if len(losses) == len(lst_th_y):
        for i, loss in enumerate(losses):
          losses_vals.append(losses[loss](th_yh[i], lst_th_y[i]))
      else:
        key = list(losses.keys())[0]
        losses_vals.append(losses[key](th_yh, lst_th_y))

    else:
      for i, loss in enumerate(losses):
        losses_vals.append(losses[loss](th_yh[i]))
    # endif
    ### TODO: sanity check
    th_loss = th.stack(losses_vals, dim=0).sum(dim=0)  # th.sum(*losses_vals, keepdim=True)

    optimizer.zero_grad()
    th_loss.mean().backward()
    optimizer.step()
    err = th_loss.detach().cpu().numpy()
    return err





import os
from datetime import datetime as dt

class _BasicPyTorchMixin(object):
  """
  Mixin for basic pytorch functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.
  """

  def __init__(self):
    super(_BasicPyTorchMixin, self).__init__()
    return

  def load_model_and_weights(
      self, model_class, weights_path, config_path,
      device='cuda', strict=True, return_config=False,
      config=None
  ):
    """
    Method for loading a model from a given class, with given configuration
    (either as a dictionary or as a .json) hyperparameters.
    Parameters
    ----------
    model_class - class extending torch.nn.Module, class of the given model
    weights_path - str, local path with the weights for the model
    config_path - str, local path with the json configuration
    device - torch.device, device on which to load the model
    strict - bool, will be used for load_state_dict()
    return_config - bool, whether to return the config as a dictionary
    config - dict, config that will be used as it is if provided

    Returns
    -------
    res - (torch.nn.Module, dict) if return_config else torch.nn.Module
    The instantiated model with the weights loaded and if required, the config dictionary of the model.
    """
    from torch import load
    try:
      model_kwargs = self.load_json(config_path) if config is None else config
      model = model_class(**model_kwargs)
      model.load_state_dict(
        load(weights_path, map_location=device),
        strict=strict
      )
      model.to(device)
    except Exception as e:
      raise e
    return (model, model_kwargs) if return_config else model

  def load_torch_weights(self, model_class, model_name, experiment_subfolder, device='cuda', strict=True):
    import torch as th
    model_files = os.listdir(
      os.path.join(
        self.get_models_folder(),
        experiment_subfolder
      )
    )
    model_files = [x for x in model_files if model_name in x]
    model_def_fn = [x for x in model_files if x.endswith('.json')][0]
    model_kwargs = self.load_json(
      os.path.join(
        self.get_models_folder(),
        experiment_subfolder,
        model_def_fn
      )
    )

    model = model_class(**model_kwargs)

    model_weights_fn = [x for x in model_files if x.endswith('.th') or x.endswith('.pth')][0]
    model.load_state_dict(th.load(
        os.path.join(self.get_models_folder(),experiment_subfolder, model_weights_fn),
        map_location=device
      ),
      strict=strict
    )

    return model

  def save_pytorch_model(self, model, label, use_prefix=False,
                         use_single_prefix=False,
                         DEBUG=True,
                         to_models=True):
    """
    DEPRECATED

    =========
    saves pytorch model to a file
    model : keras model
    label : name of the model
    use_prefix : will add date-time prefix if True (default False)
    use_single_prefix :  will use only one timestamp prefix for all saved models
    """
    import torch as th

    MDL_EXT = '.pth'
    file_prefix = ""
    if label == "":
      label = self.get_keras_model_desc(model)
    label = label.replace(">", "_")  # [:30]

    if use_single_prefix:
      file_prefix = self.file_prefix
    elif use_prefix:
      file_prefix = dt.now().strftime("%Y%m%d_%H%M%S_")

    save_folder = self.get_models_folder()
    if not to_models:
      save_folder = self.get_output_folder()

    file_name = os.path.join(save_folder, file_prefix + label)
    if file_name[-len(MDL_EXT):] == MDL_EXT:
      file_name = file_name[:-len(MDL_EXT)]
    if DEBUG:
      self.verbose_log("Saving [...{}]".format(file_name[-40:]))

    self.P("  Saving model...")
    save_file_name = file_name + MDL_EXT
    th.save(model, save_file_name)
    self.P("    Saved: {}".format(save_file_name))

    return file_name + MDL_EXT

  def load_pytorch_model(self, model_name, DEBUG=True, full_path=False):
    """
    DEPRECATED

    ===========
    Loads pytorch model

    Parameters
    ----------

    model_name : str
      The name of the model that should be loaded.

    DEBUG : boolean, optional
      Specifies whether the logging is enabled
      The default is True.

    full_path : boolean, optional
      Specifies whether `model_name` is a full path or should be loaded
      from `_models` folder.
      The default is False.
    """
    import torch as th

    name, ext = os.path.splitext(model_name)

    if ext != '.pth':
      model_name += '.pth'
    if DEBUG:
      self.verbose_log("  Trying to load {}...".format(model_name))

    if not full_path:
      model_full_path = os.path.join(self.get_models_folder(), model_name)
    else:
      model_full_path = model_name

    if os.path.isfile(model_full_path):
      if DEBUG: self.verbose_log("  Loading [...{}]".format(model_full_path[-40:]))
      model = th.load(model_full_path)
      if DEBUG: self.verbose_log("  Done loading [...{}]".format(model_full_path[-40:]), show_time=True)
    else:
      self.verbose_log("  File {} not found.".format(model_full_path))
      model = None
    return model



"""

"""
import json

from naeural_core.bc import DefaultBlockEngine
from naeural_core import Logger


if __name__ == '__main__':
  l = Logger('AENC', base_folder='.', app_folder='_local_cache')
  
  CREATE_SENDER = False
  
  if CREATE_SENDER:
    snd_name = "asym-enc-dec-1"
    sender = DefaultBlockEngine(
      log=l,
      name=snd_name,
      config= {
        "PEM_FILE"     : snd_name + ".pem",
        "PASSWORD"     : None,      
        "PEM_LOCATION" : "data"
      },
    )

  rcv_name = "asym-enc-dec-2"
  receiver = DefaultBlockEngine(
    log=l,
    name=rcv_name,
    config= {
      "PEM_FILE"     : rcv_name + ".pem",
      "PASSWORD"     : None,      
      "PEM_LOCATION" : "data"
     },
  )

  bandit = DefaultBlockEngine(
    log=l,
    name="bandit",
    config= {
      "PEM_FILE"     : "bandit.pem",
      "PASSWORD"     : None,      
      "PEM_LOCATION" : "data"
     },
  )

  l.P("*"*80)
  l.P("*"*80)
  l.P("Starting test:")

  if CREATE_SENDER:
    ## sender
    data = "Hello World"
    l.P("Sender <{}> private data: {}".format(sender.address, data))
    enc_data = sender.encrypt(
      plaintext=data,
      receiver_address=receiver.address,
    )
    l.P("Sender <{}> encrypted data: {}".format(sender.address, enc_data))
    
    payload = {
      "EE_ENCRYPTED_DATA" : True,
      "ENCRYPTED_DATA" : enc_data,    
    }
    
    signature = sender.sign(
      dct_data=payload
    )
    
    out_payload = payload
    
    str_output_payload = json.dumps(out_payload)
    print("Sender <{}> outgoing payload signed with {}:\n {}".format(
      sender.address, signature, json.dumps(out_payload, indent=2)))
  else:
    str_output_payload = """

    """

  
  ## bandit
  
  str_incoming_payload_bandit = str_output_payload
  incoming_payload_bandit = json.loads(str_incoming_payload_bandit)
  sender_address = incoming_payload_bandit.get("EE_SENDER")
  incoming_enc_data_bandit = incoming_payload_bandit.get("ENCRYPTED_DATA")

  dec_bandit = bandit.decrypt(
    encrypted_data_b64=incoming_enc_data_bandit, 
    sender_address=sender_address
  )
  l.P("Bandit <{}> decrypt attempt: {}".format(bandit.address, dec_bandit))
    
  
  ## receiver
  
  str_incoming_payload = str_output_payload
  incoming_payload = json.loads(str_incoming_payload)
  
  is_verified = receiver.verify(
    dct_data=incoming_payload
  )
    
  assert is_verified, "Payload is not valid"
  l.P("Receiver <{}> incoming payload verified: {}".format(
    receiver.address, is_verified)
  )
  
  sender_address = incoming_payload.get("EE_SENDER")  
  incomind_enc_data = incoming_payload.get("ENCRYPTED_DATA")
  
  dec = receiver.decrypt(
    encrypted_data_b64=incomind_enc_data, 
    sender_address=sender_address
  )
  l.P("Receiver <{}> decrypted message: {}".format(
    receiver.address, dec
  ))
  
  
  
  
  
  


from naeural_core import DecentrAIObject
from naeural_core import Logger

import abc
from naeural_core import constants as ct
from time import sleep
from collections import deque
from threading import Thread

class BaseHeavyOp(DecentrAIObject):

  def __init__(self, log: Logger, shmem, config, comm_async=True, **kwargs):
    self._thread = None
    self.shmem = shmem
    self.comm_async = comm_async
    # number large enough to keep all the ops, but also to ensure that memory overflow is not allowed
    self._ops = deque(maxlen=10000)
    self._config = config or {}
    self.heavy_op_count = 0
    super(BaseHeavyOp, self).__init__(log=log, **kwargs)
    return
  
  def P(self, s, color=None, **kwargs):
    if color is None or (isinstance(color,str) and color[0] not in ['e', 'r']):
      color = ct.COLORS.COMM
    super().P(s, prefix=True, color=color, **kwargs)
    return 

  def startup(self):
    super().startup()
    self.P("Initiating Heavy Ops process `{}` ".format(
      self.__class__.__name__), color='m'
    )
    self.config_data = self._config
    if self.comm_async:
      self.P("  Starting thread for `{}`".format(
        self.__class__.__name__), color='m'
      )
      self._thread = Thread(
        target=self._run_thread, 
        args=(), 
        name=ct.THREADS_PREFIX + 'hvy_' + self.__class__.__name__, 
        daemon=True,
      )
      self._thread.daemon = True
      self._thread.start()
    #endif
    return

  @property
  def cfg_idle_thread_sleep_time(self):
    return self._config.get('IDLE_THREAD_SLEEP_TIME', 1)

  @property
  def _eeid(self):
    return self.log.config_data[ct.CONFIG_STARTUP_v2.K_EE_ID][:ct.EE_ALIAS_MAX_SIZE]

  @abc.abstractmethod
  def _process_dct_operation(self, dct):
    raise NotImplementedError()

  @abc.abstractmethod
  def _register_payload_operation(self, payload):
    raise NotImplementedError()

  @staticmethod
  def __err_dict(payload, err_type, err_file, err_func, err_line, err_msg):
    return {
      'ERR_TYPE'    : err_type,
      'ERR_MSG'     : err_msg,
      'ERR_FILE'    : err_file,
      'ERR_FUNC'    : err_func,
      'ERR_LINE'    : err_line,
      'STREAM'      : payload.get('STREAM', None),
      'SIGNATURE'   : payload.get('SIGNATURE', None),
      'INSTANCE_ID' : payload.get('INSTANCE_ID', None)
    }

  def _run_thread(self):
    # thread loop for async heavy ops
    while True:
      processed = False
      if len(self._ops) > 0:
        dct = self._ops.popleft()
        self.process_dct_operation(dct)
        processed = True
      #endif

      if processed:
        sleep(0.01)
      else:
        sleep(self.cfg_idle_thread_sleep_time)
    #endwhile

  def process_dct_operation(self, dct):
    try:
      self._process_dct_operation(dct)
    except Exception as e:
      err_dict = self.__err_dict(dct, *self.log.get_error_info(return_err_val=True))
      self._create_notification(
        notif='EXCEPTION',
        msg='Error in heavy payload operation {}\n{}'.format(self.__class__.__name__, err_dict)
      )
    return

  def register_payload_operation(self, payload):
    dct = None
    try:
      dct = self._register_payload_operation(payload)
    except Exception as e:
      err_dict = self.__err_dict(payload, *self.log.get_error_info(return_err_val=True))
      self._create_notification(
        notif='EXCEPTION',
        msg='Error in heavy payload operation {}\n{}'.format(self.__class__.__name__, err_dict)
      )
    return dct

  def process_payload(self, payload):
    
    dct = self.register_payload_operation(payload)
    if self.comm_async:
      # only add to ops queue of the async thread
      self._ops.append(dct)
    else:
      # directly process the payload and eventually modify it inplace if it was 
      # correctly returned by the above `register_payload_operation`
      self.process_dct_operation(dct)
    #endif
    self.heavy_op_count += 1
    return
